# -*- coding: utf-8 -*-

"""This module contains the SQLAlchemy database models that support the definition cache and graph cache."""

import datetime
from collections import defaultdict
from typing import Any, Iterable, Mapping, Optional, Tuple

from sqlalchemy import (
    JSON,
    Boolean,
    Column,
    Date,
    DateTime,
    ForeignKey,
    Integer,
    LargeBinary,
    String,
    Table,
    Text,
    UniqueConstraint,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import backref, relationship

from .. import constants as pc
from ..constants import (
    CITATION,
    CITATION_AUTHORS,
    CITATION_DATE,
    CITATION_FIRST_AUTHOR,
    CITATION_JOURNAL,
    CITATION_LAST_AUTHOR,
    CITATION_PAGES,
    CITATION_VOLUME,
    EVIDENCE,
    IDENTIFIER,
    METADATA_AUTHORS,
    METADATA_CONTACT,
    METADATA_COPYRIGHT,
    METADATA_DESCRIPTION,
    METADATA_DISCLAIMER,
    METADATA_LICENSES,
    METADATA_NAME,
    METADATA_VERSION,
    NAME,
    NAMESPACE,
)
from ..io.gpickle import from_bytes_gz, to_bytes_gz
from ..language import CitationDict, Entity
from ..struct.graph import BELGraph
from ..tokens import parse_result_to_dsl

__all__ = [
    "Base",
    "Namespace",
    "NamespaceEntry",
    "Network",
    "Node",
    "Author",
    "Citation",
    "Evidence",
    "Edge",
    "edge_annotation",
    "network_edge",
    "network_node",
]

NAME_TABLE_NAME = "pybel_name"
NAMESPACE_TABLE_NAME = "pybel_namespace"

NODE_TABLE_NAME = "pybel_node"

EDGE_TABLE_NAME = "pybel_edge"
EDGE_ANNOTATION_TABLE_NAME = "pybel_edge_name"

AUTHOR_TABLE_NAME = "pybel_author"
CITATION_TABLE_NAME = "pybel_citation"
AUTHOR_CITATION_TABLE_NAME = "pybel_author_citation"

EVIDENCE_TABLE_NAME = "pybel_evidence"

NETWORK_TABLE_NAME = "pybel_network"
NETWORK_NODE_TABLE_NAME = "pybel_network_node"
NETWORK_EDGE_TABLE_NAME = "pybel_network_edge"
NETWORK_NAMESPACE_TABLE_NAME = "pybel_network_namespace"
NETWORK_ANNOTATION_TABLE_NAME = "pybel_network_annotation"

LONGBLOB = 4294967295

Base = declarative_base()


class Namespace(Base):
    """Represents a BEL Namespace."""

    __tablename__ = NAMESPACE_TABLE_NAME

    id = Column(Integer, primary_key=True)
    uploaded = Column(
        DateTime,
        nullable=False,
        default=datetime.datetime.utcnow,
        doc="The date of upload",
    )

    # logically the "namespace"
    keyword = Column(
        String(255),
        nullable=True,
        index=True,
        doc="Keyword that is used in a BEL file to identify a specific namespace",
    )

    # A namespace either needs a URL or a pattern
    pattern = Column(
        String(255),
        nullable=True,
        index=True,
        doc="Contains regex pattern for value identification.",
    )

    miriam_id = Column(
        String(16),
        nullable=True,
        doc=r"MIRIAM resource identifier matching the regular expression ``^MIR:001\d{5}$``",
    )
    miriam_name = Column(String(255), nullable=True)
    miriam_namespace = Column(String(255), nullable=True)
    miriam_uri = Column(String(255), nullable=True)
    miriam_description = Column(Text, nullable=True)

    version = Column(String(255), nullable=True, doc="Version of the namespace")

    url = Column(
        String(255),
        nullable=True,
        unique=True,
        index=True,
        doc="BELNS Resource location as URL",
    )

    name = Column(String(255), nullable=True, doc="Name of the given namespace")
    domain = Column(String(255), nullable=True, doc="Domain for which this namespace is valid")
    species = Column(
        String(255),
        nullable=True,
        doc="Taxonomy identifiers for which this namespace is valid",
    )
    description = Column(Text, nullable=True, doc="Optional short description of the namespace")

    created = Column(
        DateTime,
        nullable=True,
        doc="DateTime of the creation of the namespace definition file",
    )
    query_url = Column(
        Text,
        nullable=True,
        doc="URL that can be used to query the namespace (externally from PyBEL)",
    )

    author = Column(String(255), nullable=True, doc="The author of the namespace")
    license = Column(String(255), nullable=True, doc="License information")
    contact = Column(String(255), nullable=True, doc="Contact information")

    citation = Column(String(255), nullable=True)
    citation_description = Column(Text, nullable=True)
    citation_version = Column(String(255), nullable=True)
    citation_published = Column(Date, nullable=True)
    citation_url = Column(String(255), nullable=True)

    is_annotation = Column(Boolean)

    def __str__(self):
        return f"[id={self.id}] {self.keyword}"

    def get_term_to_encodings(self) -> Mapping[Tuple[Optional[str], str], str]:
        """Return the term (db, id, name) to encodings from this namespace."""
        return {(entry.identifier, entry.name): entry.encoding for entry in self.entries}

    def to_json(self, include_id: bool = False) -> Mapping[str, str]:
        """Return the most useful entries as a dictionary.

        :param include_id: If true, includes the model identifier
        """
        result = {
            "keyword": self.keyword,
            "name": self.name,
            "version": self.version,
        }

        if self.url:
            result["url"] = self.url
        else:
            result["pattern"] = self.pattern

        if include_id:
            result["id"] = self.id

        return result


class NamespaceEntry(Base):
    """Represents a name within a BEL namespace."""

    __tablename__ = NAME_TABLE_NAME
    id = Column(Integer, primary_key=True)

    name = Column(
        String(1023),
        index=True,
        nullable=True,
        doc="Name that is defined in the corresponding namespace definition file",
    )
    identifier = Column(String(255), index=True, nullable=True, doc="The database accession number")
    encoding = Column(
        String(8),
        nullable=True,
        doc="The biological entity types for which this name is valid",
    )

    namespace_id = Column(
        Integer,
        ForeignKey("{}.id".format(NAMESPACE_TABLE_NAME)),
        nullable=False,
        index=True,
    )
    namespace = relationship(Namespace, backref=backref("entries", lazy="dynamic"))

    is_name = Column(Boolean)
    is_annotation = Column(Boolean)

    def to_json(self, include_id: bool = False) -> Mapping[str, str]:
        """Describe the namespaceEntry as dictionary of Namespace-Keyword and Name.

        :param include_id: If true, includes the model identifier
        """
        result = {
            NAMESPACE: self.namespace.keyword,
        }

        if self.name:
            result[NAME] = self.name

        if self.identifier:
            result[IDENTIFIER] = self.identifier

        if include_id:
            result["id"] = self.id

        return result

    @classmethod
    def name_contains(cls, name_query: str):
        """Make a filter if the name contains a certain substring."""
        return cls.name.contains(name_query)

    def __str__(self):
        return "[id={namespace_id}] {namespace_name}:{identifier} ! {name}".format(
            namespace_id=self.namespace.id,
            namespace_name=self.namespace.keyword,
            identifier=self.identifier,
            name=self.name,
        )


network_edge = Table(
    NETWORK_EDGE_TABLE_NAME,
    Base.metadata,
    Column(
        "network_id",
        Integer,
        ForeignKey("{}.id".format(NETWORK_TABLE_NAME)),
        primary_key=True,
    ),
    Column(
        "edge_id",
        Integer,
        ForeignKey("{}.id".format(EDGE_TABLE_NAME)),
        primary_key=True,
    ),
)

network_node = Table(
    NETWORK_NODE_TABLE_NAME,
    Base.metadata,
    Column(
        "network_id",
        Integer,
        ForeignKey("{}.id".format(NETWORK_TABLE_NAME)),
        primary_key=True,
    ),
    Column(
        "node_id",
        Integer,
        ForeignKey("{}.id".format(NODE_TABLE_NAME)),
        primary_key=True,
    ),
)


class Network(Base):
    """Represents a collection of edges, specified by a BEL Script."""

    __tablename__ = NETWORK_TABLE_NAME
    id = Column(Integer, primary_key=True)

    name = Column(
        String(255),
        nullable=False,
        index=True,
        doc="Name of the given Network (from the BEL file)",
    )
    version = Column(
        String(255),
        nullable=False,
        doc="Release version of the given Network (from the BEL file)",
    )

    authors = Column(Text, nullable=True, doc="Authors of the underlying BEL file")
    contact = Column(String(255), nullable=True, doc="Contact email from the underlying BEL file")
    description = Column(Text, nullable=True, doc="Descriptive text from the underlying BEL file")
    copyright = Column(Text, nullable=True, doc="Copyright information")
    disclaimer = Column(Text, nullable=True, doc="Disclaimer information")
    licenses = Column(Text, nullable=True, doc="License information")

    created = Column(DateTime, nullable=False, default=datetime.datetime.utcnow)
    blob = Column(LargeBinary(LONGBLOB), doc="A pickled version of this network")

    nodes = relationship(
        "Node",
        secondary=network_node,
        lazy="dynamic",
        backref=backref("networks", lazy="dynamic"),
    )
    edges = relationship(
        "Edge",
        secondary=network_edge,
        lazy="dynamic",
        backref=backref("networks", lazy="dynamic"),
    )

    def to_json(self, include_id: bool = False) -> Mapping[str, Any]:
        """Return this network as JSON.

        :param include_id: If true, includes the model identifier
        """
        result = {
            METADATA_NAME: self.name,
            METADATA_VERSION: self.version,
        }

        if self.created:
            result["created"] = str(self.created)

        if include_id:
            result["id"] = self.id

        if self.authors:
            result[METADATA_AUTHORS] = self.authors

        if self.contact:
            result[METADATA_CONTACT] = self.contact

        if self.description:
            result[METADATA_DESCRIPTION] = self.description

        if self.copyright:
            result[METADATA_COPYRIGHT] = self.copyright

        if self.disclaimer:
            result[METADATA_DISCLAIMER] = self.disclaimer

        if self.licenses:
            result[METADATA_LICENSES] = self.licenses

        return result

    @classmethod
    def name_contains(cls, name_query: str):
        """Build a filter for networks whose names contain the query."""
        return cls.name.contains(name_query)

    @classmethod
    def description_contains(cls, description_query: str):
        """Build a filter for networks whose descriptions contain the query."""
        return cls.description.contains(description_query)

    @classmethod
    def id_in(cls, network_ids: Iterable[int]):
        """Build a filter for networks whose identifiers appear in the given sequence."""
        return cls.id.in_(network_ids)

    def __repr__(self):
        return "{} v{}".format(self.name, self.version)

    def __str__(self):
        return repr(self)

    def as_bel(self) -> BELGraph:
        """Get this network and loads it into a :class:`BELGraph`."""
        return from_bytes_gz(self.blob)

    def store_bel(self, graph: BELGraph):
        """Insert a BEL graph."""
        self.blob = to_bytes_gz(graph)


class Node(Base):
    """Represents a BEL Term."""

    __tablename__ = NODE_TABLE_NAME
    id = Column(Integer, primary_key=True)

    type = Column(
        String(32),
        nullable=False,
        doc="The type of the represented biological entity e.g. Protein or Gene",
    )
    bel = Column(
        String(1023),
        nullable=False,
        doc="Canonical BEL term that represents the given node",
    )
    md5 = Column(String(255), nullable=False, unique=True, index=True)

    namespace_entry_id = Column(Integer, ForeignKey("{}.id".format(NAME_TABLE_NAME)), nullable=True)
    namespace_entry = relationship(
        NamespaceEntry,
        foreign_keys=[namespace_entry_id],
        backref=backref("nodes", lazy="dynamic"),
    )

    data = Column(JSON, nullable=False, doc="PyBEL BaseEntity as JSON")

    @staticmethod
    def _start_from_base_entity(base_entity) -> "Node":
        """Convert a base entity to a node model.

        :type base_entity: pybel.dsl.BaseEntity
        """
        return Node(
            type=base_entity.function,
            bel=base_entity.as_bel(),
            md5=base_entity.md5,
            data=base_entity,
        )

    @classmethod
    def bel_contains(cls, bel_query: str):
        """Build a filter for nodes whose BEL contain the query."""
        return cls.bel.contains(bel_query)

    def __str__(self):
        return self.bel

    def __repr__(self):
        return "<Node {}: {}>".format(self.md5[:10], self.bel)

    def _get_list_by_relation(self, relation):
        return [edge.target.to_json() for edge in self.out_edges.filter(Edge.relation == relation)]

    def as_bel(self):
        """Serialize this node as a PyBEL DSL object.

        :rtype: pybel.dsl.BaseEntity
        """
        return parse_result_to_dsl(self.data)

    def to_json(self):
        """Serialize this node as a JSON object using as_bel()."""
        return self.as_bel()


author_citation = Table(
    AUTHOR_CITATION_TABLE_NAME,
    Base.metadata,
    Column(
        "author_id",
        Integer,
        ForeignKey("{}.id".format(AUTHOR_TABLE_NAME)),
        primary_key=True,
    ),
    Column(
        "citation_id",
        Integer,
        ForeignKey("{}.id".format(CITATION_TABLE_NAME)),
        primary_key=True,
    ),
)


class Author(Base):
    """Contains all author names."""

    __tablename__ = AUTHOR_TABLE_NAME

    id = Column(Integer, primary_key=True)
    name = Column(String(255), nullable=False, unique=True, index=True)

    @classmethod
    def name_contains(cls, name_query: str):
        """Build a filter for authors whose names contain the given query."""
        return cls.name.contains(name_query)

    @classmethod
    def has_name_in(cls, names: Iterable[str]):
        """Build a filter if the author has any of the given names."""
        return cls.name.in_(names)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f'Author(name="{self.name}")'


class Citation(Base):
    """The information about the citations that are used to prove a specific relation are stored in this table."""

    __tablename__ = CITATION_TABLE_NAME

    id = Column(Integer, primary_key=True)

    db = Column(String(16), nullable=False, doc="Type of the stored publication e.g. PubMed")
    db_id = Column(
        String(255),
        nullable=False,
        doc="Reference identifier of the publication e.g. PubMed_ID",
    )

    article_type = Column(Text, nullable=True, doc="Type of the publication")
    title = Column(Text, nullable=True, doc="Title of the publication")
    journal = Column(Text, nullable=True, doc="Journal name")
    volume = Column(Text, nullable=True, doc="Volume of the journal")
    issue = Column(Text, nullable=True, doc="Issue within the volume")
    pages = Column(Text, nullable=True, doc="Pages of the publication")
    date = Column(Date, nullable=True, doc="Publication date")

    first_id = Column(
        Integer,
        ForeignKey("{}.id".format(AUTHOR_TABLE_NAME)),
        nullable=True,
        doc="First author",
    )
    first = relationship(Author, foreign_keys=[first_id])

    last_id = Column(
        Integer,
        ForeignKey("{}.id".format(AUTHOR_TABLE_NAME)),
        nullable=True,
        doc="Last author",
    )
    last = relationship(Author, foreign_keys=[last_id])

    authors = relationship(Author, secondary=author_citation, backref="citations")

    __table_args__ = (UniqueConstraint(db, db_id),)

    def __str__(self):
        return "{}:{}".format(self.db, self.db_id)

    @property
    def is_pubmed(self) -> bool:
        """Return if this is a PubMed citation."""
        return self.db == "pubmed"

    @property
    def is_enriched(self) -> bool:
        """Return if this citation has been enriched for name, title, and other metadata."""
        return all(f is not None for f in (self.title, self.journal))

    def to_json(self, include_id: bool = False) -> Mapping[str, Any]:
        """Create a citation dictionary that is used to recreate the edge data dictionary of a :class:`BELGraph`.

        :param bool include_id: If true, includes the model identifier
        :return: Citation dictionary for the recreation of a :class:`BELGraph`.
        """
        result = CitationDict(
            namespace=self.db,
            identifier=self.db_id,
            name=self.title,
        )

        if include_id:
            result["id"] = self.id

        if self.title:
            result[NAME] = self.title

        if self.journal:
            result[CITATION_JOURNAL] = self.journal

        if self.volume:
            result[CITATION_VOLUME] = self.volume

        if self.pages:
            result[CITATION_PAGES] = self.pages

        if self.date:
            result[CITATION_DATE] = self.date.strftime("%Y-%m-%d")

        if self.first:
            result[CITATION_FIRST_AUTHOR] = self.first.name

        if self.last:
            result[CITATION_LAST_AUTHOR] = self.last.name

        if self.article_type:
            result[pc.CITATION_ARTICLE_TYPE] = self.article_type

        if self.authors:
            result[CITATION_AUTHORS] = sorted(author.name for author in self.authors)

        return result


class Evidence(Base):
    """This table contains the evidence text that proves a specific relationship and refers the source that is cited."""

    __tablename__ = EVIDENCE_TABLE_NAME

    id = Column(Integer, primary_key=True)
    text = Column(Text, nullable=False, doc="Supporting text from a given publication")

    citation_id = Column(Integer, ForeignKey("{}.id".format(CITATION_TABLE_NAME)), nullable=False)
    citation = relationship(Citation, backref=backref("evidences"))

    __table_args__ = (UniqueConstraint(citation_id, text),)

    def __str__(self):
        return "{}:{}:{}".format(self.citation.db, self.citation.db_id, self.text)

    def to_json(self, include_id: bool = False):
        """Create a dictionary that is used to recreate the edge data dictionary for a :class:`BELGraph`.

        :param include_id: If true, includes the model identifier
        :return: Dictionary containing citation and evidence for a :class:`BELGraph` edge.
        :rtype: dict
        """
        result = {
            CITATION: self.citation.to_json(include_id=include_id),
            EVIDENCE: self.text,
        }

        if include_id:
            result["id"] = self.id

        return result


edge_annotation = Table(
    EDGE_ANNOTATION_TABLE_NAME,
    Base.metadata,
    Column(
        "edge_id",
        Integer,
        ForeignKey("{}.id".format(EDGE_TABLE_NAME)),
        primary_key=True,
    ),
    Column(
        "name_id",
        Integer,
        ForeignKey("{}.id".format(NAME_TABLE_NAME)),
        primary_key=True,
    ),
)


class Edge(Base):
    """Relationships between BEL nodes and their properties, annotations, and provenance."""

    __tablename__ = EDGE_TABLE_NAME

    id = Column(Integer, primary_key=True)

    bel = Column(Text, nullable=False, doc="Valid BEL statement that represents the given edge")
    relation = Column(String(32), nullable=False)

    source_id = Column(Integer, ForeignKey("{}.id".format(NODE_TABLE_NAME)), nullable=False)
    source = relationship(
        Node,
        foreign_keys=[source_id],
        backref=backref("out_edges", lazy="dynamic", cascade="all, delete-orphan"),
    )

    target_id = Column(Integer, ForeignKey("{}.id".format(NODE_TABLE_NAME)), nullable=False)
    target = relationship(
        Node,
        foreign_keys=[target_id],
        backref=backref("in_edges", lazy="dynamic", cascade="all, delete-orphan"),
    )

    evidence_id = Column(Integer, ForeignKey("{}.id".format(EVIDENCE_TABLE_NAME)), nullable=True)
    evidence = relationship(Evidence, backref=backref("edges", lazy="dynamic"))

    annotations = relationship(
        NamespaceEntry,
        secondary=edge_annotation,
        lazy="dynamic",
        backref=backref("edges", lazy="dynamic"),
    )

    # free_annotations = Column(JSON, nullable=True, doc='Ungrounded extra annotations')

    source_modifier = Column(JSON, nullable=True, doc="Modifiers for the source of the edge")
    target_modifier = Column(JSON, nullable=True, doc="Modifiers for the target of the edge")

    md5 = Column(
        String(255),
        index=True,
        unique=True,
        doc="The hash of the source, target, and associated metadata",
    )

    data = Column(JSON, nullable=False, doc="The stringified JSON representing this edge")

    def __str__(self):
        return self.bel

    def __repr__(self):
        return "<Edge {}: {}>".format(self.md5, self.bel)

    def get_annotations_json(self):
        """Format the annotations properly.

        :rtype: Optional[dict[str,dict[str,bool]]
        """
        annotations = defaultdict(dict)

        for entry in self.annotations:
            annotations[entry.namespace.keyword][entry.name] = True

        return dict(annotations) or None

    def to_json(self, include_id: bool = False) -> Mapping[str, Any]:
        """Create a dictionary of one BEL Edge that can be used to create an edge in a :class:`BELGraph`.

        :param bool include_id: Include the database identifier?
        :return: Dictionary that contains information about an edge of a :class:`BELGraph`. Including participants
                 and edge data information.
        """
        source_dict = self.source.to_json()
        source_dict["md5"] = source_dict.md5
        target_dict = self.target.to_json()
        target_dict["md5"] = target_dict.md5

        result = {
            "source": source_dict,
            "target": target_dict,
            "key": self.md5,
            "data": self.data,
        }

        if include_id:
            result["id"] = self.id

        return result

    def insert_into_graph(self, graph: BELGraph) -> str:
        """Insert this edge into a BEL graph."""
        u = self.source.as_bel()
        v = self.target.as_bel()

        if self.evidence:
            return graph.add_qualified_edge(u, v, **self.data)
        else:
            return graph.add_unqualified_edge(u, v, self.relation)



# -*- coding: utf-8 -*-

"""Induction functions based on provenance information."""

import logging

from .utils import get_subgraph_by_edge_filter
from ...filters.edge_predicate_builders import (
    build_author_inclusion_filter,
    build_pmid_inclusion_filter,
)
from ...pipeline import transformation

__all__ = [
    "get_subgraph_by_pubmed",
    "get_subgraph_by_authors",
]

logger = logging.getLogger(__name__)


@transformation
def get_subgraph_by_pubmed(graph, pubmed_identifiers):
    """Induce a sub-graph over the edges retrieved from the given PubMed identifier(s).

    :param pybel.BELGraph graph: A BEL graph
    :param str or list[str] pubmed_identifiers: A PubMed identifier or list of PubMed identifiers
    :rtype: pybel.BELGraph
    """
    return get_subgraph_by_edge_filter(graph, build_pmid_inclusion_filter(pubmed_identifiers))


@transformation
def get_subgraph_by_authors(graph, authors):
    """Induce a sub-graph over the edges retrieved publications by the given author(s).

    :param pybel.BELGraph graph: A BEL graph
    :param str or list[str] authors: An author or list of authors
    :rtype: pybel.BELGraph
    """
    return get_subgraph_by_edge_filter(graph, build_author_inclusion_filter(authors))



# -*- coding: utf-8 -*-

"""Tests for :mod:`pybel.struct.summary.errors`."""

import unittest

from pybel import BELGraph
from pybel.exceptions import NakedNameWarning, UndefinedAnnotationWarning
from pybel.struct.summary import count_error_types, count_naked_names, get_naked_names
from pybel.testing.utils import n


class TestErrors(unittest.TestCase):
    """Test :mod:`pybel.struct.summary.errors`."""

    def test_count_error_types(self):
        """Test counting error types."""
        graph = BELGraph()

        line_number = 30
        position = 4
        line = n()
        annotation = n()

        exc = UndefinedAnnotationWarning(
            line_number=line_number,
            line=line,
            position=position,
            annotation=annotation,
        )

        graph.add_warning(exc)

        error_types = count_error_types(graph)

        self.assertEqual(1, len(error_types))
        self.assertIn(UndefinedAnnotationWarning.__name__, error_types)
        self.assertEqual(1, error_types[UndefinedAnnotationWarning.__name__])

    def test_get_naked_names(self):
        """Retrieve the naked names from a graph."""
        graph = BELGraph()

        n_names = 5

        line_number = 30
        position = 4
        line = n()

        names = {n() for _ in range(n_names)}

        exceptions = [
            NakedNameWarning(
                line_number=line_number,
                line=line,
                position=position,
                name=name,
            )
            for name in names
        ]

        for exception in exceptions:
            graph.add_warning(exception=exception)

        graph.add_warning(exception=exceptions[0])

        self.assertEqual(6, graph.number_of_warnings())

        naked_names = get_naked_names(graph)
        self.assertEqual(names, naked_names)

        naked_name_counter = count_naked_names(graph)
        self.assertEqual(n_names, len(naked_name_counter))
        self.assertEqual(2, naked_name_counter[exceptions[0].name])



# -*- coding: utf-8 -*-

"""Functions for inducing graphs based on edge annotations."""

import logging
from typing import Iterable, Optional, Union

from .utils import get_subgraph_by_edge_filter
from ...filters.edge_predicate_builders import (
    build_annotation_dict_all_filter,
    build_annotation_dict_any_filter,
)
from ...graph import AnnotationsHint, BELGraph
from ...pipeline import transformation

__all__ = [
    "get_subgraph_by_annotation_value",
    "get_subgraph_by_annotations",
]

logger = logging.getLogger(__name__)


@transformation
def get_subgraph_by_annotations(
    graph: BELGraph,
    annotations: AnnotationsHint,
    or_: Optional[bool] = None,
) -> BELGraph:
    """Induce a sub-graph given an annotations filter.

    :param graph: A BEL graph
    :param annotations: Annotation filters (match all with :func:`pybel.utils.subdict_matches`)
    :param or_: if True any annotation should be present, if False all annotations should be present in the
     edge. Defaults to True.
    :return: A subgraph of the original BEL graph
    """
    edge_filter_builder = build_annotation_dict_any_filter if (or_ is None or or_) else build_annotation_dict_all_filter

    annotations = graph._clean_annotations(annotations)
    return get_subgraph_by_edge_filter(graph, edge_filter_builder(annotations))


@transformation
def get_subgraph_by_annotation_value(graph: BELGraph, annotation: str, values: Union[str, Iterable[str]]) -> BELGraph:
    """Induce a sub-graph over all edges whose annotations match the given key and value.

    :param graph: A BEL graph
    :param annotation: The annotation to group by
    :param values: The value(s) for the annotation
    :return: A subgraph of the original BEL graph
    """
    if isinstance(values, str):
        values = {values}

    return get_subgraph_by_annotations(graph, {annotation: values})



# -*- coding: utf-8 -*-

"""Tests for expansion functions."""

import unittest

from pybel import BELGraph
from pybel.examples.sialic_acid_example import (
    cd33,
    cd33_phosphorylated,
    shp1,
    shp2,
    sialic_acid,
    sialic_acid_cd33_complex,
    sialic_acid_graph,
    syk,
)
from pybel.struct.mutation.expansion.neighborhood import (
    expand_node_neighborhood,
    expand_node_predecessors,
    expand_node_successors,
    expand_nodes_neighborhoods,
)


class TestExpansion(unittest.TestCase):
    """Test expansion functions."""

    def test_neighborhood(self):
        """Test expansion around the neighborhood of CD33 in the sialic acid graph given node."""
        graph = BELGraph()
        graph.add_node_from_data(cd33)
        self.assertEqual(1, graph.number_of_nodes())

        _sialic_acid_graph = sialic_acid_graph.copy()
        expand_node_neighborhood(graph=graph, universe=_sialic_acid_graph, node=cd33)

        self.assertEqual(
            {cd33, sialic_acid, sialic_acid_cd33_complex, cd33_phosphorylated},
            set(graph),
        )

    def test_neighborhood_with_predecessors(self):
        """Test expansion on the predecessors of a given node."""
        graph = BELGraph()
        graph.add_node_from_data(cd33)
        graph.add_node_from_data(sialic_acid_cd33_complex)
        self.assertEqual(3, graph.number_of_nodes())

        _sialic_acid_graph = sialic_acid_graph.copy()
        expand_node_predecessors(universe=_sialic_acid_graph, graph=graph, node=cd33)

        self.assertEqual(4, graph.number_of_nodes())
        self.assertIn(sialic_acid, graph)
        self.assertIn(sialic_acid_cd33_complex, graph)
        self.assertIn(cd33_phosphorylated, graph)

    def test_neighborhood_with_successors(self):
        """Test expansion on the successors of a given node."""
        graph = BELGraph()
        graph.add_node_from_data(cd33)
        graph.add_node_from_data(cd33_phosphorylated)
        self.assertEqual(2, graph.number_of_nodes())

        _sialic_acid_graph = sialic_acid_graph.copy()
        expand_node_successors(universe=_sialic_acid_graph, graph=graph, node=cd33)

        self.assertEqual(
            {sialic_acid_cd33_complex, sialic_acid, cd33_phosphorylated, cd33},
            set(graph),
        )

    def test_neighborhoods(self):
        """Test expansion on the neighborhood of given nodes.

        The edge between PTPN6/CD33ph should not be added.
        """
        graph = BELGraph()
        graph.add_node_from_data(cd33)
        graph.add_node_from_data(syk)
        self.assertEqual(2, graph.number_of_nodes())

        _sialic_acid_graph = sialic_acid_graph.copy()
        expand_nodes_neighborhoods(universe=_sialic_acid_graph, graph=graph, nodes=[cd33, syk])

        self.assertNotIn(shp1, graph[cd33_phosphorylated])
        self.assertNotIn(shp2, graph[cd33_phosphorylated])

        self.assertEqual(
            9,
            graph.number_of_nodes(),
            msg="wrong number of nodes: {}".format(list(graph)),
        )
        self.assertEqual(8, graph.number_of_edges(), msg="wrong number of edges")
        self.assertIn(sialic_acid, graph)
        self.assertIn(sialic_acid_cd33_complex, graph)
        self.assertIn(cd33_phosphorylated, graph)

    # TODO test that if new nodes with metadata that's missing (namespace_url definition, etc) then that gets added too



# -*- coding: utf-8 -*-

"""A simple web-based BEL parser implemented with Flask.

Run from the command line with ``python -m pybel.apps.parser``.
"""

import flask
import pyparsing

import pybel
import pybel.exceptions

app = flask.Flask(__name__)


@app.route("/parse", methods=["POST"])
def parse():
    """Parse the BEL in the `text` JSON field.

    Example usage:

    >>> import requests
    >>> requests.post('http://localhost:5000/parse', json={'text': 'p(HGNC:123) increases p(HGNC:456)'}).json()
    {'input': 'p(HGNC:123) increases p(HGNC:456)', 'output': {'object': {'concept': {'name': '456', 'namespace': 'HGNC'}, 'function': 'Protein'}, 'relation': 'increases', 'subject': {'concept': {'name': '123', 'namespace': 'HGNC'}, 'function': 'Protein'}}, 'success': True}
    """
    text = flask.request.json.get("text")
    if text is None:
        return flask.jsonify(success=False, message="missing `text`")

    try:
        rv = pybel.parse(text)
    except (pyparsing.ParseException, pybel.exceptions.PyBELWarning) as e:
        return flask.jsonify(success=False, input=text, exception=str(e))
    else:
        return flask.jsonify(success=True, input=text, output=rv)


if __name__ == "__main__":
    app.run()



# -*- coding: utf-8 -*-

"""An example with orthology statements.

The following is an example of orthology annotations from
`HomoloGene:37670 <https://www.ncbi.nlm.nih.gov/homologene?LinkName=gene_homologene&from_uid=5594>`_

.. code-block: none

    SET Citation = {"PubMed","J Immunol 1999 Sep 1 163(5) 2452-62","10452980","","",""}
    SET Evidence = "M-CSF triggers the activation of extracellular signal-regulated protein kinases (ERK)-1/2."
    SET Species = 10090
    p(MGI:Csf1) increases kin(p(MGI:Mapk1))
"""

from ..dsl import Gene, Protein, Rna, activity
from ..resources import (
    FB_URL,
    HGNC_URL,
    MGI_URL,
    NCBIGENE_URL,
    RGD_URL,
    SPECIES_PATTERN,
)
from ..struct.graph import BELGraph

__all__ = [
    "homology_graph",
]

# TODO make SGD resource

homology_graph = BELGraph(
    name="Homology and Equivalence Example Graph",
    version="1.0.1",
    description="Adds several equivalence and orthology relationships related to the mitogen-activated protein kinase "
    "(MAPK1)",
    authors="Charles Tapley Hoyt",
    contact="cthoyt@gmail.com",
)

homology_graph.namespace_url.update(
    {
        "HGNC": HGNC_URL,
        "MGI": MGI_URL,
        "RGD": RGD_URL,
        "FB": FB_URL,
        "NCBIGENE": NCBIGENE_URL,
        # 'SGD': '?',
    }
)

homology_graph.annotation_pattern.update(
    {
        "Species": SPECIES_PATTERN,
    }
)

human_mapk1_gene = Gene(namespace="HGNC", name="MAPK1", identifier="HGNC:6871")
human_mapk1_gene_entrez = Gene(namespace="NCBIGENE", name="5594")
human_mapk1_rna = Rna(namespace="HGNC", name="MAPK1", identifier="HGNC:6871")
human_mapk1_protein = Protein(namespace="HGNC", name="MAPK1", identifier="HGNC:6871")

mouse_mapk1_gene = Gene(namespace="MGI", name="Mapk1", identifier="MGI:1346858")
mouse_mapk1_gene_entrez = Gene(namespace="NCBIGENE", name="26413")
mouse_mapk1_rna = Rna(namespace="MGI", name="Mapk1", identifier="MGI:1346858")
mouse_mapk1_protein = Protein(namespace="MGI", name="Mapk1", identifier="MGI:1346858")

rat_mapk1 = Gene(namespace="RGD", name="Mapk1", identifier="70500")
rat_mapk1_entrez = Gene(namespace="NCBIGENE", name="116590")

fly_mapk1 = Gene(namespace="FB", name="rl", identifier="FBgn0003256")
fly_mapk1_entrez = Gene(namespace="NCBIGENE", name="3354888")

human_csf1_gene = Gene(namespace="HGNC", name="CSF1", identifier="HGNC:2432")
human_csf1_rna = Rna(namespace="HGNC", name="CSF1", identifier="HGNC:2432")
human_csf1_protein = Protein(namespace="HGNC", name="CSF1", identifier="HGNC:2432")

mouse_csf1_gene = Gene(namespace="MGI", name="Csf1", identifier="MGI:1339753")
mouse_csf1_rna = Rna(namespace="MGI", name="Csf1", identifier="MGI:1339753")
mouse_csf1_protein = Protein(namespace="MGI", name="Csf1", identifier="MGI:1339753")

# yeast_mapk1 = Gene(namespace='SGD', name='KSS1', identifier='SGD:S000003272')
# yeast_mapk1_entrez = Gene(namespace='NCBIGENE', name='KSS1', identifier='852931')

# TODO make homologene resource and add is_a relationships for this
# mapk1_homologene = Gene(namespace='HOMOLOGENE', identifier='37670')

homology_graph.add_equivalence(human_mapk1_gene, human_mapk1_gene_entrez)
homology_graph.add_equivalence(mouse_mapk1_gene, mouse_mapk1_gene_entrez)
homology_graph.add_equivalence(rat_mapk1, rat_mapk1_entrez)
homology_graph.add_equivalence(fly_mapk1, fly_mapk1_entrez)
# graph.add_equivalence(yeast_mapk1, yeast_mapk1_entrez)

homology_graph.add_orthology(human_csf1_gene, mouse_csf1_gene)
homology_graph.add_orthology(human_mapk1_gene, mouse_mapk1_gene)
homology_graph.add_orthology(human_mapk1_gene, rat_mapk1)
homology_graph.add_orthology(human_mapk1_gene, fly_mapk1)
# graph.add_orthology(human_mapk1, yeast_mapk1)


homology_graph.add_increases(
    source=mouse_csf1_protein,
    target=mouse_mapk1_protein,
    citation="10452980",
    evidence="M-CSF triggers the activation of extracellular signal-regulated protein kinases (ERK)-1/2.",
    target_modifier=activity("kin"),
    annotations={"Species": "10090"},
)

homology_graph.add_transcription(mouse_mapk1_gene, mouse_mapk1_rna)
homology_graph.add_translation(mouse_mapk1_rna, mouse_mapk1_protein)

homology_graph.add_transcription(human_mapk1_gene, human_mapk1_rna)
homology_graph.add_translation(human_mapk1_rna, human_mapk1_protein)

homology_graph.add_transcription(human_csf1_gene, human_csf1_rna)
homology_graph.add_translation(human_csf1_rna, human_csf1_protein)

homology_graph.add_transcription(mouse_csf1_gene, mouse_csf1_rna)
homology_graph.add_translation(mouse_csf1_rna, mouse_csf1_protein)



# -*- coding: utf-8 -*-

"""Contains the main data structure for PyBEL."""

import os
from typing import TextIO, Union

from networkx.utils import open_file
from pkg_resources import iter_entry_points

from ..struct import BELGraph

__all__ = [
    "load",
    "dump",
    "InvalidExtensionError",
]

#: Mapping from extension to importer function
IMPORTERS = {entry.name: entry.load() for entry in iter_entry_points(group="pybel.importer")}

#: Mapping from extension to exporter function
EXPORTERS = {entry.name: entry.load() for entry in iter_entry_points(group="pybel.exporter")}


class InvalidExtensionError(ValueError):
    """Raised when an invalid extension is used."""

    def __init__(self, path):
        fname = os.path.basename(path)
        super().__init__("Invalid extension for file: {}".format(fname))


def load(path: str, **kwargs) -> BELGraph:
    """Read a BEL graph.

    :param path: The path to a BEL graph in any of the formats
     with extensions described below
    :param kwargs: The keyword arguments are passed to the importer
     function
    :return: A BEL graph.

    This is the universal loader, which means any file
    path can be given and PyBEL will look up the appropriate
    load function. Allowed extensions are:

    - bel
    - bel.nodelink.json
    - bel.cx.json
    - bel.jgif.json

    The previous extensions also support gzipping.
    Other allowed extensions that don't support gzip are:

    - bel.pickle / bel.gpickle / bel.pkl
    - indra.json
    """
    for extension, importer in IMPORTERS.items():
        if path.endswith(extension):
            return importer(path, **kwargs)
    raise InvalidExtensionError(path=path)


def dump(graph: BELGraph, path: str, **kwargs) -> None:
    """Write a BEL graph.

    :param graph: A BEL graph
    :param path: The path to which the BEL graph is written.
    :param kwargs: The keyword arguments are passed to the exporter
     function

    This is the universal loader, which means any file
    path can be given and PyBEL will look up the appropriate
    writer function. Allowed extensions are:

    - bel
    - bel.nodelink.json
    - bel.unodelink.json
    - bel.cx.json
    - bel.jgif.json
    - bel.graphdati.json

    The previous extensions also support gzipping.
    Other allowed extensions that don't support gzip are:

    - bel.pickle / bel.gpickle / bel.pkl
    - indra.json
    - tsv
    - gsea
    """
    for extension, exporter in EXPORTERS.items():
        if path.endswith(extension):
            return exporter(graph, path, **kwargs)
    raise InvalidExtensionError(path=path)



# -*- coding: utf-8 -*-

"""Streamable BEL as JSON."""

import gzip
import json
from typing import Any, Iterable, List, TextIO, Union

from networkx.utils import open_file

from .nodelink import _augment_node, _prepare_graph_dict, _recover_graph_dict
from ..constants import CITATION, SOURCE_MODIFIER, TARGET_MODIFIER
from ..language import CitationDict
from ..struct.graph import BELGraph, _handle_modifier
from ..tokens import parse_result_to_dsl
from ..utils import hash_edge

__all__ = [
    "to_sbel_file",
    "to_sbel",
    "to_sbel_gz",
    "from_sbel",
    "from_sbel_gz",
    "from_sbel_file",
]

SBEL = Any


@open_file(1, mode="w")
def to_sbel_file(graph: BELGraph, path: Union[str, TextIO], separators=(",", ":"), **kwargs) -> None:
    """Write this graph as BEL JSONL to a file.

    :param graph: A BEL graph
    :param separators: The separators used in :func:`json.dumps`
    :param path: A path or file-like
    """
    for i in iterate_sbel(graph):
        print(
            json.dumps(i, ensure_ascii=False, separators=separators, **kwargs),
            file=path,
        )


def to_sbel_gz(graph: BELGraph, path: str, separators=(",", ":"), **kwargs) -> None:
    """Write a graph as BEL JSONL to a gzip file.

    :param graph: A BEL graph
    :param separators: The separators used in :func:`json.dumps`
    :param path: A path for a gzip file
    """
    with gzip.open(path, "wt") as file:
        to_sbel_file(graph, file, separators=separators, **kwargs)


def to_sbel(graph: BELGraph) -> List[SBEL]:
    """Create a list of JSON dictionaries corresponding to lines in BEL JSONL."""
    return list(iterate_sbel(graph))


def iterate_sbel(graph: BELGraph) -> Iterable[SBEL]:
    """Iterate over JSON dictionaries corresponding to lines in BEL JSONL."""
    g = graph.graph.copy()
    _prepare_graph_dict(g)
    yield g
    for u, v, k, d in graph.edges(data=True, keys=True):
        yield {
            "source": _augment_node(u),
            "target": _augment_node(v),
            "key": k,
            **d,
        }


def from_sbel(it: Iterable[SBEL], includes_metadata: bool = True) -> BELGraph:
    """Load a BEL graph from an iterable of dictionaries corresponding to lines in BEL JSONL.

    :param it: An iterable of dictionaries.
    :param includes_metadata: By default, interprets the first element of the iterable as the graph's metadata.
     Switch to ``False`` to disable.
    :return: A BEL graph
    """
    it = iter(it)
    rv = BELGraph()
    if includes_metadata:
        rv.graph.update(next(it))
        _recover_graph_dict(rv)
    add_sbel(rv, it)
    return rv


def add_sbel(graph: BELGraph, it: Iterable[SBEL]) -> None:
    """Add dictionaries to a BEL graph.

    :param graph: A BEL graph
    :param it: An iterable of dictionaries.
    """
    for data in it:
        add_sbel_row(graph, data)


def add_sbel_row(graph: BELGraph, data: SBEL) -> str:
    """Add a single SBEL data dictionary to a graph."""
    u = parse_result_to_dsl(data["source"])
    v = parse_result_to_dsl(data["target"])
    edge_data = {k: v for k, v in data.items() if k not in {"source", "target", "key"}}
    for side in (SOURCE_MODIFIER, TARGET_MODIFIER):
        side_data = edge_data.get(side)
        if side_data:
            _handle_modifier(side_data)
    if CITATION in edge_data:
        edge_data[CITATION] = CitationDict(**edge_data[CITATION])
    return graph.add_edge(u, v, key=hash_edge(u, v, edge_data), **edge_data)


@open_file(0, mode="r")
def from_sbel_file(path: Union[str, TextIO]) -> BELGraph:
    """Build a graph from the BEL JSONL contained in the given file.

    :param path: A path or file-like
    """
    return from_sbel((json.loads(line) for line in path))


def from_sbel_gz(path: str) -> BELGraph:
    """Read a graph as BEL JSONL from a gzip file."""
    with gzip.open(path, "rt") as file:
        return from_sbel_file(file)



# -*- coding: utf-8 -*-

"""Conversion functions for BEL graphs with node-link JSON."""

import gzip
import json
from io import BytesIO
from itertools import chain, count
from operator import methodcaller
from typing import Any, Mapping, TextIO, Union

from networkx.utils import open_file

from .utils import ensure_version
from ..constants import (
    ANNOTATIONS,
    CITATION,
    FUSION,
    GRAPH_ANNOTATION_CURIE,
    GRAPH_ANNOTATION_LIST,
    GRAPH_ANNOTATION_MIRIAM,
    MEMBERS,
    PARTNER_3P,
    PARTNER_5P,
    PRODUCTS,
    REACTANTS,
    SOURCE_MODIFIER,
    TARGET_MODIFIER,
)
from ..dsl import BaseEntity
from ..language import citation_dict
from ..struct import BELGraph
from ..struct.graph import _handle_modifier
from ..tokens import parse_result_to_dsl
from ..utils import hash_edge, tokenize_version

__all__ = [
    "to_nodelink",
    "to_nodelink_file",
    "to_nodelink_gz",
    "to_nodelink_jsons",
    "from_nodelink",
    "from_nodelink_file",
    "from_nodelink_gz",
    "from_nodelink_jsons",
    "to_nodelink_gz_io",
    "from_nodelink_gz_io",
]


def to_nodelink(graph: BELGraph) -> Mapping[str, Any]:
    """Convert this graph to a node-link JSON object.

    :param graph: BEL Graph
    """
    graph_json_dict = _to_nodelink_json_helper(graph)

    _prepare_graph_dict(graph_json_dict["graph"])

    return graph_json_dict


def _prepare_graph_dict(g):
    # Convert annotation list definitions (which are sets) to canonicalized/sorted lists
    g[GRAPH_ANNOTATION_LIST] = {
        keyword: list(sorted(values)) for keyword, values in g.get(GRAPH_ANNOTATION_LIST, {}).items()
    }

    g[GRAPH_ANNOTATION_CURIE] = list(sorted(g[GRAPH_ANNOTATION_CURIE]))
    g[GRAPH_ANNOTATION_MIRIAM] = list(sorted(g[GRAPH_ANNOTATION_MIRIAM]))


@open_file(1, mode="w")
def to_nodelink_file(graph: BELGraph, path: Union[str, TextIO], **kwargs) -> None:
    """Write this graph as node-link JSON to a file.

    :param graph: A BEL graph
    :param path: A path or file-like
    """
    graph_json_dict = to_nodelink(graph)
    json.dump(graph_json_dict, path, ensure_ascii=False, **kwargs)


def to_nodelink_gz(graph, path: str, **kwargs) -> None:
    """Write a graph as node-link JSON to a gzip file."""
    with gzip.open(path, "wt") as file:
        json.dump(to_nodelink(graph), file, ensure_ascii=False, **kwargs)


def to_nodelink_jsons(graph: BELGraph, **kwargs) -> str:
    """Dump this graph as a node-link JSON object to a string."""
    return json.dumps(to_nodelink(graph), ensure_ascii=False, **kwargs)


def from_nodelink(graph_json_dict: Mapping[str, Any], check_version: bool = True) -> BELGraph:
    """Build a graph from node-link JSON Object."""
    pybel_version = tokenize_version(graph_json_dict["graph"]["pybel_version"])
    if pybel_version[1] < 14:  # if minor version is less than 14
        raise ValueError("Invalid NodeLink JSON from old version of PyBEL (v{}.{}.{})".format(*pybel_version))
    graph = _from_nodelink_json_helper(graph_json_dict)
    return ensure_version(graph, check_version=check_version)


@open_file(0, mode="r")
def from_nodelink_file(path: Union[str, TextIO], check_version: bool = True) -> BELGraph:
    """Build a graph from the node-link JSON contained in the given file.

    :param path: A path or file-like
    """
    return from_nodelink(json.load(path), check_version=check_version)


def from_nodelink_gz(path: str) -> BELGraph:
    """Read a graph as node-link JSON from a gzip file."""
    with gzip.open(path, "rt") as file:
        return from_nodelink(json.load(file))


def from_nodelink_jsons(graph_json_str: str, check_version: bool = True) -> BELGraph:
    """Read a BEL graph from a node-link JSON string."""
    return from_nodelink(json.loads(graph_json_str), check_version=check_version)


def _to_nodelink_json_helper(graph: BELGraph) -> Mapping[str, Any]:
    """Convert a BEL graph to a node-link format.

    :param graph: BEL Graph

    Adapted from :func:`networkx.readwrite.json_graph.node_link_data`
    """
    nodes = sorted(graph, key=methodcaller("as_bel"))

    mapping = dict(zip(nodes, count()))

    return {
        "directed": True,
        "multigraph": True,
        "graph": graph.graph.copy(),
        "nodes": [_augment_node(node) for node in nodes],
        "links": [
            dict(
                chain(
                    data.copy().items(),
                    [("source", mapping[u]), ("target", mapping[v]), ("key", key)],
                ),
            )
            for u, v, key, data in graph.edges(keys=True, data=True)
        ],
    }


def _augment_node(node: BaseEntity) -> BaseEntity:
    """Add the SHA-512 identifier to a node's dictionary."""
    rv = node.copy()
    rv["id"] = node.md5
    rv["bel"] = node.as_bel()
    for m in chain(node.get(MEMBERS, []), node.get(REACTANTS, []), node.get(PRODUCTS, [])):
        m.update(_augment_node(m))
    if FUSION in node:
        node[FUSION][PARTNER_3P].update(_augment_node(node[FUSION][PARTNER_3P]))
        node[FUSION][PARTNER_5P].update(_augment_node(node[FUSION][PARTNER_5P]))
    return rv


def _recover_graph_dict(graph: BELGraph):
    graph.graph[GRAPH_ANNOTATION_LIST] = {
        keyword: set(values) for keyword, values in graph.graph.get(GRAPH_ANNOTATION_LIST, {}).items()
    }
    graph.graph[GRAPH_ANNOTATION_CURIE] = set(graph.graph.get(GRAPH_ANNOTATION_CURIE, []))
    graph.graph[GRAPH_ANNOTATION_MIRIAM] = set(graph.graph.get(GRAPH_ANNOTATION_MIRIAM, []))


def _from_nodelink_json_helper(data: Mapping[str, Any]) -> BELGraph:
    """Return graph from node-link data format.

    Adapted from :func:`networkx.readwrite.json_graph.node_link_graph`
    """
    graph = BELGraph()
    graph.graph = data.get("graph", {})
    _recover_graph_dict(graph)

    mapping = []

    for node_data in data["nodes"]:
        node = parse_result_to_dsl(node_data)
        graph.add_node_from_data(node)
        mapping.append(node)

    for data in data["links"]:
        u = mapping[data["source"]]
        v = mapping[data["target"]]

        edge_data = {k: v for k, v in data.items() if k not in {"source", "target", "key"}}

        for side in (SOURCE_MODIFIER, TARGET_MODIFIER):
            side_data = edge_data.get(side)
            if side_data:
                _handle_modifier(side_data)

        if CITATION in edge_data:
            edge_data[CITATION] = citation_dict(**edge_data[CITATION])

        if ANNOTATIONS in edge_data:
            edge_data[ANNOTATIONS] = graph._clean_annotations(edge_data[ANNOTATIONS])

        graph.add_edge(u, v, key=hash_edge(u, v, edge_data), **edge_data)

    return graph


def to_nodelink_gz_io(graph: BELGraph) -> BytesIO:
    """Get a BEL graph as a compressed BytesIO."""
    bytes_io = BytesIO()
    with gzip.GzipFile(fileobj=bytes_io, mode="w") as file:
        s = to_nodelink_jsons(graph)
        file.write(s.encode("utf-8"))
    bytes_io.seek(0)
    return bytes_io


def from_nodelink_gz_io(bytes_io: BytesIO) -> BELGraph:
    """Get BEL from gzipped nodelink JSON."""
    with gzip.GzipFile(fileobj=bytes_io, mode="r") as file:
        s = file.read()
    j = s.decode("utf-8")
    return from_nodelink_jsons(j)



# -*- coding: utf-8 -*-

"""Functions for selecting by the neighborhoods of nodes."""

import itertools as itt
from typing import Iterable, Optional

from ...graph import BELGraph
from ...pipeline import transformation
from ...utils import update_metadata
from ....dsl import BaseEntity

__all__ = [
    "get_subgraph_by_neighborhood",
]


@transformation
def get_subgraph_by_neighborhood(graph: BELGraph, nodes: Iterable[BaseEntity]) -> Optional[BELGraph]:
    """Get a BEL graph around the neighborhoods of the given nodes.

    Returns none if no nodes are in the graph.

    :param graph: A BEL graph
    :param nodes: An iterable of BEL nodes
    :return: A BEL graph induced around the neighborhoods of the given nodes
    """
    node_set = set(nodes)

    if not any(node in graph for node in node_set):
        return

    rv = graph.child()
    rv.add_edges_from(
        itt.chain(
            graph.in_edges(nodes, keys=True, data=True),
            graph.out_edges(nodes, keys=True, data=True),
        ),
    )
    return rv



# -*- coding: utf-8 -*-

"""Utilities for the PyBEL database manager."""

from typing import Dict, Mapping, Optional, Tuple, Union

from ..utils import parse_datetime


def extract_shared_required(config, definition_header: str = "Namespace"):
    """Get the required annotations shared by BEL namespace and annotation resource documents.

    :param dict config: The configuration dictionary representing a BEL resource
    :param definition_header: ``Namespace`` or ``AnnotationDefinition``
    :rtype: dict
    """
    return {
        "keyword": config[definition_header]["Keyword"],
        "created": parse_datetime(config[definition_header]["CreatedDateTime"]),
    }


def extract_shared_optional(bel_resource, definition_header: str = "Namespace"):
    """Get the optional annotations shared by BEL namespace and annotation resource documents.

    :param dict bel_resource: A configuration dictionary representing a BEL resource
    :param definition_header: ``Namespace`` or ``AnnotationDefinition``
    :rtype: dict
    """
    shared_mapping = {
        "description": (definition_header, "DescriptionString"),
        "version": (definition_header, "VersionString"),
        "author": ("Author", "NameString"),
        "license": ("Author", "CopyrightString"),
        "contact": ("Author", "ContactInfoString"),
        "citation": ("Citation", "NameString"),
        "citation_description": ("Citation", "DescriptionString"),
        "citation_version": ("Citation", "PublishedVersionString"),
        "citation_url": ("Citation", "ReferenceURL"),
    }

    result = {}

    update_insert_values(bel_resource, shared_mapping, result)

    if "PublishedDate" in bel_resource.get("Citation", {}):
        result["citation_published"] = parse_datetime(bel_resource["Citation"]["PublishedDate"])

    return result


def update_insert_values(
    bel_resource: Mapping,
    mapping: Mapping[str, Tuple[str, str]],
    values: Dict[str, str],
) -> None:
    """Update the value dictionary with a BEL resource dictionary."""
    for database_column, (section, key) in mapping.items():
        if section in bel_resource and key in bel_resource[section]:
            values[database_column] = bel_resource[section][key]


def int_or_str(v: Optional[str]) -> Union[None, int, str]:
    """Safe converts an string represent an integer to an integer or passes through ``None``."""
    if v is None:
        return
    try:
        return int(v)
    except ValueError:
        return v



# -*- coding: utf-8 -*-

"""This module contains helper functions for reading BEL scripts."""

import logging
import os
import re
import time
from typing import Any, Iterable, List, Mapping, Optional, Tuple

from bel_resources import ResourceError, split_file_to_annotations_and_definitions
from pyparsing import ParseException
from sqlalchemy.exc import OperationalError
from tqdm.autonotebook import tqdm

from ..constants import INVERSE_DOCUMENT_KEYS, REQUIRED_METADATA
from ..exceptions import (
    BELParserWarning,
    BELSyntaxError,
    InconsistentDefinitionError,
    MalformedMetadataException,
    MissingMetadataException,
    PlaceholderAminoAcidWarning,
    VersionFormatWarning,
)
from ..manager import Manager
from ..parser import BELParser, MetadataParser
from ..struct.graph import BELGraph

__all__ = [
    "parse_lines",
]

logger = logging.getLogger(__name__)
parser_logger = logging.getLogger("pybel.parser")

METADATA_LINE_RE = re.compile(r"(SET\s+DOCUMENT|DEFINE\s+NAMESPACE|DEFINE\s+ANNOTATION)")
LOG_FMT = "%d:%d %s %s"
LOG_FMT_PATH = "%s:%d:%d %s %s"


def parse_lines(
    graph: BELGraph,
    lines: Iterable[str],
    manager: Optional[Manager] = None,
    disallow_nested: bool = False,
    citation_clearing: bool = True,
    use_tqdm: bool = False,
    tqdm_kwargs: Optional[Mapping[str, Any]] = None,
    no_identifier_validation: bool = False,
    disallow_unqualified_translocations: bool = False,
    allow_redefinition: bool = False,
    allow_definition_failures: bool = False,
    allow_naked_names: bool = False,
    required_annotations: Optional[List[str]] = None,
    upgrade_urls: bool = False,
) -> None:
    """Parse an iterable of lines into this graph.

    Delegates to :func:`parse_document`, :func:`parse_definitions`, and :func:`parse_statements`.

    :param graph: A BEL graph
    :param lines: An iterable over lines of BEL script
    :param manager: A PyBEL database manager
    :param disallow_nested: If true, turns on nested statement failures
    :param citation_clearing: Should :code:`SET Citation` statements clear evidence and all annotations?
                                   Delegated to :class:`pybel.parser.ControlParser`
    :param use_tqdm: Use :mod:`tqdm` to show a progress bar?
    :param tqdm_kwargs: Keywords to pass to ``tqdm``
    :param disallow_unqualified_translocations: If true, allow translocations without TO and FROM clauses.
    :param required_annotations: Annotations that are required for all statements
    :param upgrade_urls: Automatically upgrade old namespace URLs. Defaults to false.

    .. warning::

        These options allow concessions for parsing BEL that is either **WRONG** or **UNSCIENTIFIC**. Use them at
        risk to reproducibility and validity of your results.

    :param no_identifier_validation: If true, turns off namespace validation
    :param allow_naked_names: If true, turns off naked namespace failures
    :param allow_redefinition: If true, doesn't fail on second definition of same name or annotation
    :param allow_definition_failures: If true, allows parsing to continue if a terminology file download/parse fails
    """
    docs, definitions, statements = split_file_to_annotations_and_definitions(lines)

    if manager is None:
        manager = Manager()

    metadata_parser = MetadataParser(
        manager,
        allow_redefinition=allow_redefinition,
        skip_validation=no_identifier_validation,
        upgrade_urls=upgrade_urls,
    )

    parse_document(
        graph,
        docs,
        metadata_parser,
    )

    parse_definitions(
        graph,
        definitions,
        metadata_parser,
        allow_failures=allow_definition_failures,
        use_tqdm=use_tqdm,
        tqdm_kwargs=tqdm_kwargs,
    )

    bel_parser = BELParser(
        graph=graph,
        # terminologies
        namespace_to_term_to_encoding=metadata_parser.namespace_to_term_to_encoding,
        namespace_to_pattern=metadata_parser.namespace_to_pattern,
        annotation_to_term=metadata_parser.annotation_to_term,
        annotation_to_pattern=metadata_parser.annotation_to_pattern,
        annotation_to_local=metadata_parser.annotation_to_local,
        # language settings
        disallow_nested=disallow_nested,
        citation_clearing=citation_clearing,
        skip_validation=no_identifier_validation,
        allow_naked_names=allow_naked_names,
        disallow_unqualified_translocations=disallow_unqualified_translocations,
        required_annotations=required_annotations,
    )

    parse_statements(
        graph,
        statements,
        bel_parser,
        use_tqdm=use_tqdm,
        tqdm_kwargs=tqdm_kwargs,
    )

    logger.info(
        "Network has %d nodes and %d edges",
        graph.number_of_nodes(),
        graph.number_of_edges(),
    )


def parse_document(
    graph: BELGraph,
    enumerated_lines: Iterable[Tuple[int, str]],
    metadata_parser: MetadataParser,
) -> None:
    """Parse the lines in the document section of a BEL script."""
    parse_document_start_time = time.time()

    for line_number, line in enumerated_lines:
        try:
            metadata_parser.parseString(line, line_number=line_number)
        except VersionFormatWarning as exc:
            _log_parse_exception(graph, exc)
            graph.add_warning(exc)
        except Exception as e:
            exc = MalformedMetadataException(line_number, line, 0)
            _log_parse_exception(graph, exc)
            raise exc from e

    for required in REQUIRED_METADATA:
        required_metadatum = metadata_parser.document_metadata.get(required)
        if required_metadatum is not None:
            continue

        required_metadatum_key = INVERSE_DOCUMENT_KEYS[required]
        # This has to be insert since it needs to go on the front!
        exc = MissingMetadataException.make(required_metadatum_key)
        graph.warnings.insert(0, (None, exc, {}))
        _log_parse_exception(graph, exc)

    graph.document.update(metadata_parser.document_metadata)

    logger.info(
        "Finished parsing document section in %.02f seconds",
        time.time() - parse_document_start_time,
    )


def parse_definitions(
    graph: BELGraph,
    enumerated_lines: Iterable[Tuple[int, str]],
    metadata_parser: MetadataParser,
    allow_failures: bool = False,
    use_tqdm: bool = False,
    tqdm_kwargs: Optional[Mapping[str, Any]] = None,
) -> None:
    """Parse the lines in the definitions section of a BEL script.

    :param graph: A BEL graph
    :param enumerated_lines: An enumerated iterable over the lines in the definitions section of a BEL script
    :param metadata_parser: A metadata parser
    :param allow_failures: If true, allows parser to continue past strange failures
    :param use_tqdm: Use :mod:`tqdm` to show a progress bar?
    :param tqdm_kwargs: Keywords to pass to ``tqdm``
    :raises: pybel.parser.parse_exceptions.InconsistentDefinitionError
    :raises: pybel.resources.exc.ResourceError
    :raises: sqlalchemy.exc.OperationalError
    """
    parse_definitions_start_time = time.time()

    if use_tqdm:
        _tqdm_kwargs = dict(desc="Definitions", leave=False)
        if tqdm_kwargs:
            _tqdm_kwargs.update(tqdm_kwargs)
        enumerated_lines = tqdm(list(enumerated_lines), **_tqdm_kwargs)

    for line_number, line in enumerated_lines:
        try:
            metadata_parser.parseString(line, line_number=line_number)
        except (InconsistentDefinitionError, ResourceError) as e:
            parser_logger.exception(LOG_FMT, line_number, 0, e.__class__.__name__, line)
            raise e
        except OperationalError as e:
            parser_logger.warning(
                "Need to upgrade database. See http://pybel.readthedocs.io/en/latest/installation.html#upgrading",
            )
            raise e
        except Exception as e:
            if not allow_failures:
                exc = MalformedMetadataException(line_number, line, 0)
                _log_parse_exception(graph, exc)
                raise exc from e

    graph.namespace_url.update(metadata_parser.namespace_url_dict)
    graph.namespace_pattern.update(
        {keyword: pattern.pattern for keyword, pattern in metadata_parser.namespace_to_pattern.items()}
    )

    graph.annotation_url.update(metadata_parser.annotation_url_dict)
    graph.annotation_pattern.update(
        {keyword: pattern.pattern for keyword, pattern in metadata_parser.annotation_to_pattern.items()}
    )
    graph.annotation_list.update(metadata_parser.annotation_to_local)

    logger.info(
        "Finished parsing definitions section in %.02f seconds",
        time.time() - parse_definitions_start_time,
    )

    metadata_parser.ensure_resources()
    logger.info("Finished ensuring namespaces in cache")


def parse_statements(
    graph: BELGraph,
    enumerated_lines: Iterable[Tuple[int, str]],
    bel_parser: BELParser,
    use_tqdm: bool = True,
    tqdm_kwargs: Optional[Mapping[str, Any]] = None,
) -> None:
    """Parse a list of statements from a BEL Script.

    :param graph: A BEL graph
    :param enumerated_lines: An enumerated iterable over the lines in the statements section of a BEL script
    :param bel_parser: A BEL parser
    :param use_tqdm: Use :mod:`tqdm` to show a progress bar? Requires reading whole file to memory.
    :param tqdm_kwargs: Keywords to pass to ``tqdm``
    """
    parse_statements_start_time = time.time()

    if use_tqdm:
        tqdm_kwargs = {} if tqdm_kwargs is None else dict(tqdm_kwargs)
        tqdm_kwargs.setdefault("desc", "Statements")
        tqdm_kwargs.setdefault("leave", False)
        enumerated_lines = tqdm(list(enumerated_lines), **tqdm_kwargs)

    for line_number, line in enumerated_lines:
        try:
            bel_parser.parseString(line, line_number=line_number)
        except ParseException as e:
            exc = BELSyntaxError(line_number, line, e.loc)
            _log_parse_exception(graph, exc)
            graph.add_warning(exc, bel_parser.get_annotations())
        except PlaceholderAminoAcidWarning as exc:
            exc.line_number = line_number
            _log_parse_exception(graph, exc)
            graph.add_warning(exc, bel_parser.get_annotations())
        except BELParserWarning as exc:
            _log_parse_exception(graph, exc)
            graph.add_warning(exc, bel_parser.get_annotations())
        except Exception:
            parser_logger.exception(LOG_FMT, line_number, 0, "General Failure", line)
            raise

    logger.info(
        "Parsed statements section in %.02f seconds with %d warnings",
        time.time() - parse_statements_start_time,
        len(graph.warnings),
    )


def _log_parse_exception(graph: BELGraph, exc: BELParserWarning):
    if graph.path:
        s = LOG_FMT_PATH % (
            os.path.basename(graph.path),
            exc.line_number,
            exc.position,
            exc.__class__.__name__,
            exc,
        )
    else:
        s = LOG_FMT % (exc.line_number, exc.position, exc.__class__.__name__, exc)
    tqdm.write(s)



# -*- coding: utf-8 -*-

"""Tests for the jsonschema node validation."""

import copy
import json
import os
import unittest

import pybel.dsl
from pybel.schema import is_valid_edge, is_valid_node
from pybel.testing.utils import n

NAMESPACE, NAME, IDENTIFIER = n(), n(), n()

BLANK_ABUNDANCE = {"function": "", "variants": []}

# Abundances
PROTEIN = pybel.dsl.Protein(namespace=NAMESPACE, name=NAME, identifier=IDENTIFIER)
GENE = pybel.dsl.Gene(NAMESPACE, name=NAME, identifier=IDENTIFIER)

# Variants
GENE_MOD = pybel.dsl.GeneModification(NAME, namespace=NAMESPACE, identifier=IDENTIFIER)
PROTEIN_SUB = pybel.dsl.ProteinSubstitution("Ala", 1, "Tyr")
PROTEIN_MOD = pybel.dsl.ProteinModification(NAME, code="Ala", position=1, namespace=NAMESPACE, identifier=IDENTIFIER)
FRAGMENT = pybel.dsl.Fragment(1, 2)


class TestNodeSchema(unittest.TestCase):
    """Tests for the jsonschema node validation."""

    def test_pure_abundances(self):
        """Test validating abundance nodes."""
        self.assertTrue(is_valid_node(PROTEIN))
        self.assertTrue(is_valid_node(GENE))

        abundance = pybel.dsl.Abundance(namespace=NAMESPACE, name=NAME, identifier=IDENTIFIER)
        self.assertTrue(is_valid_node(abundance))

    def test_variant_abundances(self):
        """Test validating abundance nodes with variants."""
        gmod = GENE.with_variants(GENE_MOD)
        self.assertTrue(is_valid_node(gmod))

        psub = PROTEIN.with_variants(PROTEIN_SUB)
        self.assertTrue(is_valid_node(psub))

        pmod = PROTEIN.with_variants(PROTEIN_MOD)
        self.assertTrue(is_valid_node(pmod))

        frag = PROTEIN.with_variants(FRAGMENT)
        self.assertTrue(is_valid_node(frag))

    def test_matching_variants(self):
        """Test catching invalid abundance/variant pairs, e.g. ProteinModification on a Gene."""
        invalid_gmod = PROTEIN.with_variants(GENE_MOD)
        self.assertFalse(is_valid_node(invalid_gmod))

        invalid_psub = GENE.with_variants(PROTEIN_SUB)
        self.assertFalse(is_valid_node(invalid_psub))

        invalid_pmod = GENE.with_variants(PROTEIN_MOD)
        self.assertFalse(is_valid_node(invalid_pmod))

        invalid_frag = GENE.with_variants(FRAGMENT)
        self.assertFalse(is_valid_node(invalid_frag))

    def test_fusions(self):
        """Test validating fusion nodes."""
        protein = pybel.dsl.Protein(namespace=NAMESPACE, name=NAME, identifier=IDENTIFIER)
        enum_fusion_range = pybel.dsl.EnumeratedFusionRange("r", 1, 79)
        missing_fusion_range = pybel.dsl.MissingFusionRange()
        fusion_node = pybel.dsl.FusionBase(protein, protein, range_5p=enum_fusion_range, range_3p=missing_fusion_range)
        self.assertTrue(is_valid_node(fusion_node))

    def test_list_abundances(self):
        """Test validating list abundance nodes."""
        complex_abundance = pybel.dsl.ComplexAbundance(
            [GENE.with_variants(GENE_MOD), PROTEIN.with_variants(PROTEIN_SUB)],
            namespace=NAMESPACE,
            name=NAME,
            identifier=IDENTIFIER,
        )
        composite_abundance = pybel.dsl.CompositeAbundance([PROTEIN, complex_abundance])
        self.assertTrue(is_valid_node(complex_abundance))
        self.assertTrue(is_valid_node(composite_abundance))

    def test_reaction(self):
        """Test validating a reaction node."""
        node = pybel.dsl.Rna(namespace=NAMESPACE, name=NAME, identifier=IDENTIFIER)
        node_list = [node, PROTEIN, GENE]
        rxn = pybel.dsl.Reaction(reactants=node_list, products=node_list)
        self.assertTrue(is_valid_node(rxn))

    def test_invalid_abundances(self):
        """Test that invalid abundances are caught."""
        missing_fn = BLANK_ABUNDANCE.copy()
        missing_fn.pop("function")
        self.assertFalse(is_valid_node(missing_fn))

    def test_invalid_psub(self):
        """Test that invalid protein substitutions are caught."""
        missing_hgvs = dict(PROTEIN_SUB)
        # Remove the required "hgvs" property
        missing_hgvs.pop("hgvs")
        protein = BLANK_ABUNDANCE.copy()
        protein["function"] = "Protein"
        protein["variants"] = [missing_hgvs]
        self.assertFalse(is_valid_node(protein))

    def test_invalid_amino(self):
        """Test that protein variants with invalid amino acid codes are caught."""
        invalid_psub = dict(PROTEIN_SUB)
        invalid_psub["hgvs"] = "p.Aaa0Bbb"
        self.assertFalse(is_valid_node(invalid_psub))

        invalid_pmod = dict(PROTEIN_MOD)
        invalid_pmod["code"] = "Aaa"
        self.assertFalse(is_valid_node(invalid_pmod))


class TestEdgeSchema(unittest.TestCase):
    """Tests for the jsonschema edge validation."""

    @classmethod
    def setUpClass(cls):
        """Load the edge contained in example_edge.json."""
        here = os.path.abspath(os.path.dirname(__file__))
        example_file = os.path.join(here, "example_edge.json")
        with open(example_file) as example_json:
            edge = json.load(example_json)
        cls.example_edge = edge

    def test_predefined_example(self):
        """Test a predefined edge example."""
        edge = self.example_edge
        self.assertTrue(is_valid_edge(edge))

    def test_missing_information(self):
        """Test removing information from the predefined edge."""
        edge = self.example_edge

        missing_source = copy.deepcopy(edge)
        missing_source.pop("source")
        self.assertFalse(is_valid_edge(missing_source))

        missing_relation = copy.deepcopy(edge)
        missing_relation.pop("relation")
        self.assertFalse(is_valid_edge(missing_relation))

        missing_target = copy.deepcopy(edge)
        missing_target.pop("target")
        self.assertFalse(is_valid_edge(missing_target))

        missing_location = copy.deepcopy(edge)
        missing_location["target"]["effect"].pop("fromLoc")
        self.assertFalse(is_valid_edge(missing_location))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

"""Test for functions for inducing random sub-graphs."""

import random
import sys
import unittest
from collections import Counter

import networkx as nx

from pybel.examples import sialic_acid_graph, statin_graph
from pybel.struct.mutation.induction.paths import get_random_path
from pybel.struct.mutation.induction.random_subgraph import (
    _helper,
    get_graph_with_random_edges,
    get_random_node,
    get_random_subgraph,
)
from pybel.testing.generate import generate_random_graph


@unittest.skipIf(sys.version_info < (3,), "Will not support random operations on python2")
class TestRandom(unittest.TestCase):
    """Test random graph induction functions."""

    def setUp(self):
        """Set the random seed before each test."""
        random.seed(125)  # love that number

    def test_random_edges(self):
        """Test getting a graph by random edges."""
        n_nodes, n_edges, n_sample_edges = 15, 80, 40
        graph = generate_random_graph(n_nodes=n_nodes, n_edges=n_edges)

        subgraph = get_graph_with_random_edges(graph, n_edges=n_sample_edges)
        self.assertEqual(n_sample_edges, subgraph.number_of_edges())

    def test_random_nodes(self):
        """Test getting random nodes."""
        graph = nx.MultiDiGraph()

        graph.add_edge(1, 2)
        graph.add_edge(1, 3)
        graph.add_edge(1, 4)
        graph.add_edge(1, 5)

        n = 30000
        r = Counter(get_random_node(graph, set(), invert_degrees=False) for _ in range(n))

        degree_sum = 4 + 1 + 1 + 1 + 1

        self.assertAlmostEqual(4 / degree_sum, r[1] / n, places=2)
        self.assertAlmostEqual(1 / degree_sum, r[2] / n, places=2)
        self.assertAlmostEqual(1 / degree_sum, r[3] / n, places=2)
        self.assertAlmostEqual(1 / degree_sum, r[4] / n, places=2)
        self.assertAlmostEqual(1 / degree_sum, r[5] / n, places=2)

    def test_random_nodes_inverted(self):
        """Test getting random nodes."""
        graph = nx.MultiDiGraph()

        graph.add_edge(1, 2)
        graph.add_edge(1, 3)
        graph.add_edge(1, 4)
        graph.add_edge(1, 5)

        n = 30000
        r = Counter(get_random_node(graph, set(), invert_degrees=True) for _ in range(n))

        degree_sum = (1 / 4) + (1 / 1) + (1 / 1) + (1 / 1) + (1 / 1)

        self.assertAlmostEqual((1 / 4) / degree_sum, r[1] / n, places=2)
        self.assertAlmostEqual((1 / 1) / degree_sum, r[2] / n, places=2)
        self.assertAlmostEqual((1 / 1) / degree_sum, r[3] / n, places=2)
        self.assertAlmostEqual((1 / 1) / degree_sum, r[4] / n, places=2)
        self.assertAlmostEqual((1 / 1) / degree_sum, r[5] / n, places=2)

    def test_random_sample(self):
        """Test randomly sampling a graph."""
        n_nodes, n_edges = 50, 500
        graph = generate_random_graph(n_nodes=n_nodes, n_edges=n_edges)

        self.assertEqual(n_edges, graph.number_of_edges())

        sg_1 = get_random_subgraph(graph, number_edges=250, number_seed_edges=10, invert_degrees=False)
        self.assertEqual(250, sg_1.number_of_edges())

        sg_2 = get_random_subgraph(graph, number_edges=250, number_seed_edges=10, invert_degrees=True)
        self.assertEqual(250, sg_2.number_of_edges())

    def test_random_sample_small(self):
        """Test a graph that is too small to sample."""
        n_nodes, n_edges = 11, 25
        graph = generate_random_graph(n_nodes, n_edges)

        self.assertEqual(n_edges, graph.number_of_edges())

        sg_1 = get_random_subgraph(graph, number_edges=250, number_seed_edges=5, invert_degrees=False)
        self.assertEqual(
            graph.number_of_edges(),
            sg_1.number_of_edges(),
            msg="since graph is too small, the subgraph should contain the whole thing",
        )

        sg_2 = get_random_subgraph(graph, number_edges=250, number_seed_edges=5, invert_degrees=True)
        self.assertEqual(
            graph.number_of_edges(),
            sg_2.number_of_edges(),
            msg="since graph is too small, the subgraph should contain the whole thing",
        )

    def test_helper_failure(self):
        graph = nx.MultiDiGraph()
        graph.add_edge(1, 2)
        graph.add_edge(2, 3)

        result = nx.MultiDiGraph()
        result.add_edge(1, 2)

        _helper(
            result,
            graph,
            number_edges_remaining=5,
            node_blacklist={1, 2, 3},
        )

        self.assertNotIn(3, result)


class TestRandomPath(unittest.TestCase):
    """Test getting random paths."""

    def test_get_random_path(self):
        """Test getting random paths doesn't crash."""
        for graph in (sialic_acid_graph, statin_graph):
            for _ in range(100):
                get_random_path(graph.copy())



# -*- coding: utf-8 -*-

"""This module contains helper functions for other IO functions."""

from .exc import ImportVersionWarning
from ..config import PYBEL_MINIMUM_IMPORT_VERSION
from ..struct import BELGraph
from ..utils import tokenize_version


def raise_for_old_graph(graph):
    """Raise an ImportVersionWarning if the BEL graph was produced by a legacy version of PyBEL.

    :param graph: A BEL graph
    :raises ImportVersionWarning: If the BEL graph was produced by a legacy version of PyBEL
    """
    graph_version = tokenize_version(graph.pybel_version)
    if graph_version < PYBEL_MINIMUM_IMPORT_VERSION:
        raise ImportVersionWarning(graph_version, PYBEL_MINIMUM_IMPORT_VERSION)


def raise_for_not_bel(graph):
    """Raise a TypeError if the argument is not a BEL graph.

    :param graph: A BEL graph
    :raises TypeError: If the argument is not a BEL graph
    """
    if not isinstance(graph, BELGraph):
        raise TypeError("Not a BELGraph: {}".format(graph))


def ensure_version(graph: BELGraph, check_version: bool = True) -> BELGraph:
    """Ensure that the graph was produced by a minimum of PyBEL v:data:`PYBEL_MINIMUM_IMPORT_VERSION`.

    This variable is defined by last release with a change in the graph data definition.

    :param graph: A BEL Graph
    :param check_version: Should the version be checked, or should the graph just be returned without inspection
    :raises ImportVersionWarning: If the BEL graph was produced by a legacy version of PyBEL
    """
    if check_version:
        raise_for_old_graph(graph)

    return graph



# -*- coding: utf-8 -*-

"""Protein Substitutions.

Protein substitutions are legacy statements defined in BEL 1.0. BEL 2.0 recommends using HGVS strings. Luckily,
the information contained in a BEL 1.0 encoding, such as :code:`p(HGNC:APP,sub(R,275,H))` can be
automatically translated to the appropriate HGVS :code:`p(HGNC:APP, var(p.Arg275His))`, assuming that all
substitutions are using the reference protein sequence for numbering and not the genomic reference.
The previous statements both produce the underlying data:

.. code-block:: python

    from pybel.constants import *

    {
        FUNCTION: GENE,
        NAMESPACE: 'HGNC',
        NAME: 'APP',
        VARIANTS: [
            {
                KIND: HGVS,
                IDENTIFIER: 'p.Arg275His',
            },
        ],
    }

.. seealso::

    - BEL 2.0 specification on `protein substitutions
      <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_variants_2>`_
    - PyBEL module :py:class:`pybel.parser.modifiers.get_protein_substitution_language`
"""

import logging

from pyparsing import ParserElement
from pyparsing import pyparsing_common as ppc

from .constants import amino_acid
from ..utils import nest, one_of_tags
from ...constants import HGVS, KIND, PSUB_POSITION, PSUB_REFERENCE, PSUB_VARIANT

__all__ = [
    "get_protein_substitution_language",
]

logger = logging.getLogger(__name__)

psub_tag = one_of_tags(tags=["sub", "substitution"], canonical_tag=HGVS, name=KIND)


def get_protein_substitution_language() -> ParserElement:
    """Build a protein substitution parser."""
    parser_element = psub_tag + nest(
        amino_acid(PSUB_REFERENCE),
        ppc.integer(PSUB_POSITION),
        amino_acid(PSUB_VARIANT),
    )
    parser_element.setParseAction(_handle_psub)
    return parser_element


def _handle_psub(line, _, tokens):
    upgraded = "p.{}{}{}".format(tokens[PSUB_REFERENCE], tokens[PSUB_POSITION], tokens[PSUB_VARIANT])
    logger.log(5, "sub() in p() is deprecated: %s. Upgraded to %s", line, upgraded)
    tokens[HGVS] = upgraded
    del tokens[PSUB_REFERENCE]
    del tokens[PSUB_POSITION]
    del tokens[PSUB_VARIANT]
    return tokens



# -*- coding: utf-8 -*-

"""Utilities for PyBEL."""

import hashlib
import json
import logging
import re
import typing
from collections import defaultdict
from collections.abc import Iterable, MutableMapping
from datetime import datetime
from typing import Any, List, Mapping, Optional, Tuple, TypeVar

from .constants import (
    ACTIVITY,
    CITATION,
    DEGRADATION,
    EFFECT,
    EVIDENCE,
    FROM_LOC,
    IDENTIFIER,
    LOCATION,
    MODIFIER,
    NAME,
    NAMESPACE,
    RELATION,
    SOURCE_MODIFIER,
    TARGET_MODIFIER,
    TO_LOC,
    TRANSLOCATION,
)
from .typing import EdgeData

try:
    import pickle5 as pickle
except ImportError:
    import pickle

logger = logging.getLogger(__name__)

CanonicalEdge = Tuple[str, Optional[Tuple], Optional[Tuple]]


def expand_dict(flat_dict, sep: str = "_"):
    """Expand a flattened dictionary.

    :param dict flat_dict: a nested dictionary that has been flattened so the keys are composite
    :param sep: the separator between concatenated keys
    :rtype: dict
    """
    res = {}
    rdict = defaultdict(list)

    for flat_key, value in flat_dict.items():
        key = flat_key.split(sep, 1)
        if 1 == len(key):
            res[key[0]] = value
        else:
            rdict[key[0]].append((key[1:], value))

    for k, v in rdict.items():
        res[k] = expand_dict({ik: iv for (ik,), iv in v})

    return res


def flatten_dict(
    data: Mapping[str, Any],
    parent_key: str = "",
    sep: str = "_",
) -> Mapping[str, str]:
    """Flatten a nested dictionary.

    :param data: A nested dictionary
    :param parent_key: The parent's key. This is a value for tail recursion, so don't set it yourself.
    :param sep: The separator used between dictionary levels
    :rtype: dict

    .. seealso:: http://stackoverflow.com/a/6027615
    """
    items = {}

    for key, value in data.items():
        # prepend the parent key
        key = parent_key + sep + key if parent_key else key

        if isinstance(value, (dict, MutableMapping)):
            items.update(flatten_dict(value, key, sep=sep))
        elif isinstance(value, (set, list)):
            items[key] = ",".join(value)
        else:
            items[key] = value

    return items


def tokenize_version(version_string: str) -> Tuple[int, int, int]:
    """Tokenize a version string to a tuple.

    Truncates qualifiers like ``-dev``.

    :param version_string: A version string
    :return: A tuple representing the version string

    >>> tokenize_version('0.1.2-dev')
    (0, 1, 2)
    """
    before_dash = version_string.split("-")[0]
    major, minor, patch = before_dash.split(".")[:3]  # take only the first 3 in case there's an extension like -dev.0
    return int(major), int(minor), int(patch)


_re = re.compile(r"^[a-zA-Z0-9-\.]*$")


def ensure_quotes(s: str) -> str:
    """Quote a string that isn't solely alphanumeric."""
    s.isalnum()
    return s if _re.match(s) else f'"{s}"'


CREATION_DATE_FMT = "%Y-%m-%dT%H:%M:%S"
PUBLISHED_DATE_FMT = "%Y-%m-%d"
PUBLISHED_DATE_FMT_2 = "%d:%m:%Y %H:%M"
DATE_VERSION_FMT = "%Y%m%d"


def valid_date(s: str) -> bool:
    """Check that a string represents a valid date in ISO 8601 format YYYY-MM-DD."""
    return _validate_date_fmt(s, PUBLISHED_DATE_FMT)


def valid_date_version(s: str) -> bool:
    """Check that the string is a valid date versions string."""
    return _validate_date_fmt(s, DATE_VERSION_FMT)


def _validate_date_fmt(s: str, fmt: str) -> bool:
    try:
        datetime.strptime(s, fmt)
    except ValueError:
        return False
    else:
        return True


def parse_datetime(s: str) -> datetime.date:
    """Try to parse a datetime object from a standard datetime format or date format."""
    for fmt in (CREATION_DATE_FMT, PUBLISHED_DATE_FMT, PUBLISHED_DATE_FMT_2):
        try:
            dt = datetime.strptime(s, fmt)
        except ValueError:
            pass
        else:
            return dt

    raise ValueError("Incorrect datetime format for {}".format(s))


def _get_citation_str(data: Mapping) -> Optional[str]:
    citation = data.get(CITATION)
    if citation is not None:
        return citation.curie


def hash_edge(source, target, edge_data: EdgeData) -> str:
    """Convert an edge tuple to a MD5 hash.

    :param BaseEntity source: The source BEL node
    :param BaseEntity target: The target BEL node
    :param edge_data: The edge's data dictionary
    :return: A hashed version of the edge tuple using MD5 hash of the binary pickle dump of u, v, and the json dump
     of d
    """
    edge_tuple = _get_edge_tuple(source, target, edge_data)
    edge_tuple_bytes = pickle.dumps(edge_tuple)
    return hashlib.md5(edge_tuple_bytes).hexdigest()  # noqa: S303


def _get_edge_tuple(
    source,
    target,
    edge_data: EdgeData,
) -> Tuple[str, str, Optional[str], Optional[str], CanonicalEdge]:
    """Convert an edge to a consistent tuple.

    :param BaseEntity source: The source BEL node
    :param BaseEntity target: The target BEL node
    :param edge_data: The edge's data dictionary
    :return: A tuple that can be hashed representing this edge. Makes no promises to its structure.
    """
    return (
        source.as_bel(),
        target.as_bel(),
        _get_citation_str(edge_data),
        edge_data.get(EVIDENCE),
        canonicalize_edge(edge_data),
    )


def subdict_matches(target: Mapping, query: Mapping, partial_match: bool = True) -> bool:
    """Check if all the keys in the query dict are in the target dict, and that their values match.

    1. Checks that all keys in the query dict are in the target dict
    2. Matches the values of the keys in the query dict
        a. If the value is a string, then must match exactly
        b. If the value is a set/list/tuple, then will match any of them
        c. If the value is a dict, then recursively check if that subdict matches

    :param target: The dictionary to search
    :param query: A query dict with keys to match
    :param partial_match: Should the query values be used as partial or exact matches? Defaults to :code:`True`.
    :return: if all keys in b are in target_dict and their values match
    """
    for k, v in query.items():
        if k not in target:
            return False
        elif not isinstance(v, (int, str, dict, Iterable)):
            raise ValueError("invalid value: {}".format(v))
        elif isinstance(v, (int, str)) and target[k] != v:
            return False
        elif isinstance(v, dict):
            if partial_match:
                if not isinstance(target[k], dict):
                    return False
                elif not subdict_matches(target[k], v, partial_match):
                    return False
            elif not partial_match and target[k] != v:
                return False
        elif isinstance(v, Iterable) and target[k] not in v:
            return False

    return True


def hash_dump(data) -> str:
    """Hash an arbitrary JSON dictionary by dumping it in sorted order, encoding it in UTF-8, then hashing the bytes.

    :param data: An arbitrary JSON-serializable object
    :type data: dict or list or tuple
    """
    return hashlib.md5(json.dumps(data, sort_keys=True).encode("utf-8")).hexdigest()  # noqa: S303


def canonicalize_edge(edge_data: EdgeData) -> CanonicalEdge:
    """Canonicalize the edge to a tuple based on the relation, subject modifications, and object modifications."""
    return (
        edge_data[RELATION],
        _canonicalize_edge_modifications(edge_data.get(SOURCE_MODIFIER)),
        _canonicalize_edge_modifications(edge_data.get(TARGET_MODIFIER)),
    )


def _canonicalize_edge_modifications(edge_data: EdgeData) -> Optional[Tuple]:
    """Return the SUBJECT or OBJECT entry of a PyBEL edge data dictionary as a canonical tuple."""
    if edge_data is None:
        return

    modifier = edge_data.get(MODIFIER)
    location = edge_data.get(LOCATION)
    effect = edge_data.get(EFFECT)

    if modifier is None and location is None:
        return

    result = []

    if modifier == ACTIVITY:
        if effect:
            t = (
                ACTIVITY,
                effect[NAMESPACE],
                effect.get(IDENTIFIER),
                effect.get(NAME),
            )
        else:
            t = (ACTIVITY,)
        result.append(t)

    elif modifier == DEGRADATION:
        t = (DEGRADATION,)
        result.append(t)

    elif modifier == TRANSLOCATION:
        if effect:
            from_loc_concept = effect[FROM_LOC]
            to_loc_concept = effect[TO_LOC]

            t = (
                TRANSLOCATION,
                from_loc_concept[NAMESPACE],
                from_loc_concept.get(IDENTIFIER),
                from_loc_concept.get(NAME),
                to_loc_concept[NAMESPACE],
                to_loc_concept.get(IDENTIFIER),
                to_loc_concept.get(NAME),
            )
        else:
            t = (TRANSLOCATION,)
        result.append(t)

    if location:
        t = (
            LOCATION,
            location[NAMESPACE],
            location.get(IDENTIFIER),
            location.get(NAME),
        )
        result.append(t)

    if not result:
        raise ValueError("Invalid data: {}".format(edge_data))

    return tuple(result)


def get_corresponding_pickle_path(path: str) -> str:
    """Get the same path with a pickle extension.

    :param path: A path to a BEL file.
    """
    return "{path}.pickle".format(path=path)


X = TypeVar("X")
Y = TypeVar("Y")


def multidict(pairs: typing.Iterable[Tuple[X, Y]]) -> Mapping[X, List[Y]]:
    """Accumulate a multidict from a list of pairs."""
    rv = defaultdict(list)
    for key, value in pairs:
        rv[key].append(value)
    return dict(rv)



# -*- coding: utf-8 -*-

"""A wrapper around selection methods."""

import logging
from typing import Any, List, Optional

from .constants import (
    SEED_TYPE_ANNOTATION,
    SEED_TYPE_AUTHOR,
    SEED_TYPE_DOUBLE_NEIGHBORS,
    SEED_TYPE_DOWNSTREAM,
    SEED_TYPE_INDUCTION,
    SEED_TYPE_NEIGHBORS,
    SEED_TYPE_PATHS,
    SEED_TYPE_PUBMED,
    SEED_TYPE_SAMPLE,
    SEED_TYPE_UPSTREAM,
)
from ..mutation import (
    expand_nodes_neighborhoods,
    get_multi_causal_downstream,
    get_multi_causal_upstream,
    get_random_subgraph,
    get_subgraph_by_all_shortest_paths,
    get_subgraph_by_annotations,
    get_subgraph_by_authors,
    get_subgraph_by_induction,
    get_subgraph_by_neighborhood,
    get_subgraph_by_pubmed,
    get_subgraph_by_second_neighbors,
)
from ...dsl import BaseEntity

__all__ = [
    "get_subgraph",
]

logger = logging.getLogger(__name__)


def get_subgraph(
    graph,
    seed_method: Optional[str] = None,
    seed_data: Optional[Any] = None,
    expand_nodes: Optional[List[BaseEntity]] = None,
    remove_nodes: Optional[List[BaseEntity]] = None,
):
    """Run a pipeline query on graph with multiple sub-graph filters and expanders.

    Order of Operations:

    1. Seeding by given function name and data
    2. Add nodes
    3. Remove nodes

    :param pybel.BELGraph graph: A BEL graph
    :param seed_method: The name of the get_subgraph_by_* function to use
    :param seed_data: The argument to pass to the get_subgraph function
    :param expand_nodes: Add the neighborhoods around all of these nodes
    :param remove_nodes: Remove these nodes and all of their in/out edges
    :rtype: Optional[pybel.BELGraph]
    """
    # Seed by the given function
    if seed_method == SEED_TYPE_INDUCTION:
        result = get_subgraph_by_induction(graph, seed_data)

    elif seed_method == SEED_TYPE_PATHS:
        result = get_subgraph_by_all_shortest_paths(graph, seed_data)

    elif seed_method == SEED_TYPE_NEIGHBORS:
        result = get_subgraph_by_neighborhood(graph, seed_data)

    elif seed_method == SEED_TYPE_DOUBLE_NEIGHBORS:
        result = get_subgraph_by_second_neighbors(graph, seed_data)

    elif seed_method == SEED_TYPE_UPSTREAM:
        result = get_multi_causal_upstream(graph, seed_data)

    elif seed_method == SEED_TYPE_DOWNSTREAM:
        result = get_multi_causal_downstream(graph, seed_data)

    elif seed_method == SEED_TYPE_PUBMED:
        result = get_subgraph_by_pubmed(graph, seed_data)

    elif seed_method == SEED_TYPE_AUTHOR:
        result = get_subgraph_by_authors(graph, seed_data)

    elif seed_method == SEED_TYPE_ANNOTATION:
        result = get_subgraph_by_annotations(graph, seed_data["annotations"], or_=seed_data.get("or"))

    elif seed_method == SEED_TYPE_SAMPLE:
        result = get_random_subgraph(
            graph,
            number_edges=seed_data.get("number_edges"),
            seed=seed_data.get("seed"),
        )

    elif not seed_method:  # Otherwise, don't seed a sub-graph
        result = graph.copy()
        logger.debug("no seed function - using full network: %s", result.name)

    else:
        raise ValueError("Invalid seed method: {}".format(seed_method))

    if result is None:
        logger.debug("query returned no results")
        return

    logger.debug(
        "original graph has (%s nodes / %s edges)",
        result.number_of_nodes(),
        result.number_of_edges(),
    )

    # Expand around the given nodes
    if expand_nodes:
        expand_nodes_neighborhoods(graph, result, expand_nodes)
        logger.debug(
            "graph expanded to (%s nodes / %s edges)",
            result.number_of_nodes(),
            result.number_of_edges(),
        )

    # Delete the given nodes
    if remove_nodes:
        for node in remove_nodes:
            if node not in result:
                logger.debug("%s is not in graph %s", node, graph.name)
                continue
            result.remove_node(node)
        logger.debug(
            "graph contracted to (%s nodes / %s edges)",
            result.number_of_nodes(),
            result.number_of_edges(),
        )

    logger.debug(
        "Subgraph coming from %s (seed type) %s (data) contains %d nodes and %d edges",
        seed_method,
        seed_data,
        result.number_of_nodes(),
        result.number_of_edges(),
    )

    return result



# -*- coding: utf-8 -*-

"""Ecosystem of Machine-maintained Models with Automated Analysis (EMMAA).

`EMMAA <https://emmaa.indra.bio/>`_ is a project built on top of INDRA by
the Sorger Lab at Harvard Medical School. It automatically builds knowledge
graphs around pathways/indications periodically (almost daily) using the
INDRA Database, which in turn is updated periodically (almost daily)
with the most recent literature from MEDLINE, PubMed Central, several
major publishers, and other bespoke text corpora such as CORD-19.
"""

import json
import logging
from typing import Iterable, Optional
from xml.etree import ElementTree  # noqa:S405

import click
import requests
from more_click import verbose_option

from .indra import from_indra_statements
from ..struct import BELGraph

__all__ = [
    "from_emmaa",
]

logger = logging.getLogger(__name__)


def from_emmaa(
    model: str,
    *,
    date: Optional[str] = None,
    extension: Optional[str] = None,
    suppress_warnings: bool = False,
) -> BELGraph:
    """Get an EMMAA model as a BEL graph.

    Get the most recent COVID-19 model from EMMAA with the following:

    .. code-block:: python

        import pybel

        covid19_emmaa_graph = pybel.from_emmaa('covid19', extension='jsonl')
        covid19_emmaa_graph.summarize()

    PyBEL does its best to look up the most recent model, but if that doesn't work,
    you can specify it explicitly with the ``date`` keyword argument in the form
    of ``%Y-%m-%d-%H-%M-%S`` like in the following:

    .. code-block:: python

        import pybel

        covid19_emmaa_graph = pybel.from_emmaa('covid19', '2020-04-23-17-44-57', extension='jsonl')
        covid19_emmaa_graph.summarize()
    """
    statements = get_statements_from_emmaa(
        model=model,
        date=date,
        extension=extension,
        suppress_warnings=suppress_warnings,
    )
    return from_indra_statements(statements, name=model, version=date)


def get_statements_from_emmaa(
    model: str,
    *,
    date: Optional[str] = None,
    extension: Optional[str] = None,
    suppress_warnings: bool = False,
):
    """Get INDRA statements from EMMAA.

    :rtype: List[indra.statements.Statement]
    """
    from indra.statements import stmts_from_json

    if suppress_warnings:
        logging.getLogger("indra.assemblers.pybel.assembler").setLevel(logging.ERROR)
        logging.getLogger("indra.sources.bel.processor").setLevel(logging.ERROR)
    if extension is None:
        extension = "json"

    if date is None:
        url = f"https://emmaa.s3.amazonaws.com/assembled/{model}/latest_statements_{model}.{extension}"
    else:
        url = f"https://emmaa.s3.amazonaws.com/assembled/{model}/statements_{date}.{extension}"

    res = requests.get(url)

    if extension == "jsonl":
        res_json = [json.loads(line) for line in res.text.splitlines()]
        return stmts_from_json(res_json)
    elif extension == "json":
        res_json = res.json()
        return stmts_from_json(res_json)
    elif extension == "gz":
        raise NotImplementedError
    else:
        raise ValueError(f"unhandled extension: {extension}")


def _get_latest_date(model: str) -> str:
    res = requests.get("https://emmaa.s3.amazonaws.com/")
    tree = ElementTree.fromstring(res.text)  # noqa:S314
    return max(_iter_dates(tree, model))


def _iter_dates(tree: ElementTree, model: str) -> Iterable[str]:
    aws = "{http://s3.amazonaws.com/doc/2006-03-01/}"
    for x in tree.findall(f"{aws}Contents/{aws}Key"):
        prefix = f"assembled/{model}/statements_"
        if x.text.startswith(prefix):
            yield x.text


@click.command()
@verbose_option
@click.option("--extension", default="jsonl")
def main(extension: str):
    """Run the EMMAA converter."""
    from_emmaa("covid19", extension=extension).summarize()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Truncations.

Truncations in the legacy BEL 1.0 specification are automatically translated to BEL 2.0 with HGVS nomenclature.
:code:`p(HGNC:AKT1, trunc(40))` becomes :code:`p(HGNC:AKT1, var(p.40*))` and is represented with the following
dictionary:

.. code-block:: python

    from pybel.constants import *

    {
        FUNCTION: PROTEIN,
        NAMESPACE: 'HGNC',
        NAME: 'AKT1',
        VARIANTS: [
            {
                KIND: HGVS,
                IDENTIFIER: 'p.40*',
            },
        ],
    }

Unfortunately, the HGVS nomenclature requires the encoding of the terminal amino acid which is exchanged
for a stop codon, and this information is not required by BEL 1.0. For this example, the proper encoding
of the truncation at position also includes the information that the 40th amino acid in the AKT1 is Cys. Its
BEL encoding should be :code:`p(HGNC:AKT1, var(p.Cys40*))`. Temporary support has been added to
compile these statements, but it's recommended they are upgraded by reexamining the supporting text, or
looking up the amino acid sequence.

.. seealso::

    - BEL 2.0 specification on `truncations <http://openbel.org/language/version_2.0/bel_specification_version_2.0.html#_variants_2>`_
    - PyBEL module :py:class:`pybel.parser.modifiers.get_truncation_language`
"""

import logging

from pyparsing import ParserElement
from pyparsing import pyparsing_common as ppc

from .constants import amino_acid
from ..utils import nest, one_of_tags
from ...constants import HGVS, KIND, TRUNCATION_POSITION

__all__ = [
    "get_truncation_language",
]

logger = logging.getLogger(__name__)

truncation_tag = one_of_tags(tags=["trunc", "truncation"], canonical_tag=HGVS, name=KIND)

AMINO_ACID = "aminoacid"


def get_truncation_language() -> ParserElement:
    """Build a parser for protein truncations."""
    l1 = truncation_tag + nest(amino_acid(AMINO_ACID) + ppc.integer(TRUNCATION_POSITION))
    l1.setParseAction(_handle_trunc)
    l2 = truncation_tag + nest(ppc.integer(TRUNCATION_POSITION))
    l2.setParseAction(_handle_trunc_legacy)
    return l1 | l2


def _handle_trunc_legacy(line, _, tokens):
    # FIXME this isn't correct HGVS nomenclature, but truncation isn't forward compatible without more information
    upgraded = "p.{}*".format(tokens[TRUNCATION_POSITION])
    logger.warning(
        "trunc() is deprecated. Re-encode with reference terminal amino acid in HGVS: %s",
        line,
    )
    tokens[HGVS] = upgraded
    del tokens[TRUNCATION_POSITION]
    return tokens


def _handle_trunc(_, __, tokens):
    aa, position = tokens[AMINO_ACID], tokens[TRUNCATION_POSITION]
    tokens[HGVS] = "p.{aa}{position}*".format(aa=aa, position=position)
    del tokens[AMINO_ACID]
    del tokens[TRUNCATION_POSITION]
    return tokens



# -*- coding: utf-8 -*-

"""Type hints for the parsers."""

from typing import Mapping, Optional, Tuple

__all__ = [
    "Term",
    "TermEncodingMapping",
    "NamespaceTermEncodingMapping",
]

Term = Tuple[Optional[str], str]
TermEncodingMapping = Mapping[Term, str]
NamespaceTermEncodingMapping = Mapping[str, Mapping[Term, str]]



# -*- coding: utf-8 -*-

"""Query builder for PyBEL."""

from .exc import *
from .query import Query
from .seeding import SEED_DATA, SEED_METHOD, Seeding
from .selection import get_subgraph

__all__ = [k for k in locals() if not k.startswith("_")]



# -*- coding: utf-8 -*-

import unittest
from typing import Set, Tuple

from pybel import BELGraph
from pybel.constants import ANNOTATIONS
from pybel.dsl import BaseEntity, Protein
from pybel.struct.filters import (
    and_edge_predicates,
    concatenate_node_predicates,
    count_passed_edge_filter,
    count_passed_node_filter,
    filter_edges,
    get_nodes,
    invert_edge_predicate,
)
from pybel.struct.filters.edge_predicate_builders import (
    _annotation_dict_all_filter,
    _annotation_dict_any_filter,
    build_annotation_dict_all_filter,
    build_annotation_dict_any_filter,
)
from pybel.struct.filters.edge_predicates import true_edge_predicate
from pybel.struct.filters.node_predicates import true_node_predicate
from pybel.struct.filters.typing import EdgeIterator
from pybel.testing.utils import n


def make_edge_iterator_set(it: EdgeIterator) -> Set[Tuple[BaseEntity, BaseEntity]]:
    return {(u, v) for u, v, _ in it}


class TestNodeFilters(unittest.TestCase):
    def setUp(self):
        self.universe = BELGraph()

        self.universe.add_edge(1, 2)
        self.universe.add_edge(2, 3)
        self.universe.add_edge(3, 7)
        self.universe.add_edge(1, 4)
        self.universe.add_edge(1, 5)
        self.universe.add_edge(5, 6)
        self.universe.add_edge(8, 2)

        self.graph = BELGraph()
        self.graph.add_edge(1, 2)

        self.all_universe_nodes = {1, 2, 3, 4, 5, 6, 7, 8}
        self.all_graph_nodes = {1, 2}

    def test_no_node_filter_argument(self):
        nodes = get_nodes(self.universe, [])
        self.assertEqual(self.all_universe_nodes, nodes)

    def test_keep_node_permissive(self):
        nodes = get_nodes(self.universe, true_node_predicate)
        self.assertEqual(self.all_universe_nodes, nodes)

    def test_missing_node_filter(self):
        nodes = get_nodes(self.universe, concatenate_node_predicates([]))
        self.assertEqual(self.all_universe_nodes, nodes)

    def test_concatenate_single_node_filter(self):
        nodes = get_nodes(self.universe, [true_node_predicate])
        self.assertEqual(self.all_universe_nodes, nodes)

    def test_concatenate_multiple_node_filters(self):
        def even(_, node) -> bool:
            return node % 2 == 0

        def big(_, node) -> bool:
            return node > 3

        nodes = get_nodes(self.universe, [even, big])
        self.assertEqual({4, 6, 8}, nodes)

        self.assertEqual(3, count_passed_node_filter(self.universe, [even, big]))

    def test_no_edge_filter(self):
        edges = make_edge_iterator_set(filter_edges(self.graph, []))
        self.assertEqual({(1, 2)}, edges)

    def test_keep_edge_permissive(self):
        edges = make_edge_iterator_set(filter_edges(self.graph, true_edge_predicate))
        self.assertEqual({(1, 2)}, edges)

    def test_keep_edge_unpermissive(self):
        keep_edge_restrictive = invert_edge_predicate(true_edge_predicate)
        edges = make_edge_iterator_set(filter_edges(self.graph, keep_edge_restrictive))
        self.assertEqual(set(), edges)

    def test_missing_edge_filter(self):
        edges = make_edge_iterator_set(filter_edges(self.graph, and_edge_predicates([])))
        self.assertEqual(({(1, 2)}), edges)

    def test_concatenate_single_edge_filter(self):
        edges = make_edge_iterator_set(filter_edges(self.graph, [true_edge_predicate]))
        self.assertEqual({(1, 2)}, edges)

    def test_concatenate_multiple_edge_filter(self):
        def has_odd_source(graph, u, v, k):
            return u % 2 != 0

        def has_even_target(graph, u, v, k):
            return v % 2 == 0

        edges = make_edge_iterator_set(filter_edges(self.universe, [has_odd_source, has_even_target]))
        self.assertEqual({(1, 2), (1, 4), (5, 6)}, edges)

        self.assertEqual(
            3,
            count_passed_edge_filter(self.universe, [has_odd_source, has_even_target]),
        )

        has_even_source = invert_edge_predicate(has_odd_source)
        edges = make_edge_iterator_set(filter_edges(self.universe, has_even_source))
        self.assertEqual({(2, 3), (8, 2)}, edges)


class TestEdgeFilters(unittest.TestCase):
    def test_a(self):
        self.assertTrue(_annotation_dict_any_filter({ANNOTATIONS: {"A": {"1", "2"}}}, {"A": {"1"}}))

        self.assertTrue(_annotation_dict_any_filter({ANNOTATIONS: {"A": {"1", "2"}}}, {"A": {"1", "2"}}))

        self.assertTrue(_annotation_dict_any_filter({ANNOTATIONS: {"A": {"1", "2"}}}, {"A": {"1", "2", "3"}}))

        self.assertTrue(
            _annotation_dict_any_filter({ANNOTATIONS: {"A": {"1", "2"}, "B": {"X"}}}, {"A": {"3"}, "B": {"X"}})
        )

        self.assertFalse(_annotation_dict_any_filter({ANNOTATIONS: {"A": {"1", "2"}}}, {"A": {"3"}}))

        self.assertFalse(
            _annotation_dict_any_filter({ANNOTATIONS: {"A": {"1", "2"}, "B": {"X"}}}, {"A": {"3"}, "B": {"Y"}})
        )

    def test_any_filter_no_query(self):
        """Test that the all filter returns true when there's no argument"""
        graph = BELGraph()
        graph.add_increases(Protein(n(), n()), Protein(n(), n()), citation=n(), evidence=n())
        self.assertEqual(1, count_passed_edge_filter(graph, build_annotation_dict_any_filter({})))

    def test_any_filter_no_annotations(self):
        graph = BELGraph()
        graph.add_increases(Protein(n(), n()), Protein(n(), n()), citation=n(), evidence=n())
        self.assertEqual(
            0,
            count_passed_edge_filter(graph, build_annotation_dict_any_filter({"A": {"1"}})),
        )

    def test_any_filter_empty_annotations(self):
        graph = BELGraph()
        graph.add_increases(
            Protein(n(), n()),
            Protein(n(), n()),
            citation=n(),
            evidence=n(),
            annotations={},
        )
        self.assertEqual(
            0,
            count_passed_edge_filter(graph, build_annotation_dict_any_filter({"A": {"1"}})),
        )

    def test_any_filter(self):
        graph = BELGraph()
        graph.annotation_list["A"] = set("12345")
        graph.add_increases(
            Protein(n(), n()),
            Protein(n(), n()),
            citation=n(),
            evidence=n(),
            annotations={"A": {"1", "2", "3"}},
        )

        self.assertEqual(
            1,
            count_passed_edge_filter(
                graph,
                build_annotation_dict_any_filter(graph._clean_annotations({"A": {"1"}})),
            ),
        )
        self.assertEqual(
            1,
            count_passed_edge_filter(
                graph,
                build_annotation_dict_any_filter(graph._clean_annotations({"A": {"1", "2"}})),
            ),
        )
        self.assertEqual(
            1,
            count_passed_edge_filter(
                graph,
                build_annotation_dict_any_filter(graph._clean_annotations({"A": {"1", "2", "3"}})),
            ),
        )

    def test_b(self):
        self.assertTrue(_annotation_dict_all_filter({ANNOTATIONS: {"A": {"1"}}}, {"A": {"1"}}))

        self.assertTrue(_annotation_dict_all_filter({ANNOTATIONS: {"A": {"1", "2"}}}, {"A": {"1", "2"}}))

        self.assertTrue(_annotation_dict_all_filter({ANNOTATIONS: {"A": {"1", "2"}}}, {"A": {"1", "2"}}))

        self.assertTrue(
            _annotation_dict_all_filter(
                {ANNOTATIONS: {"A": {"1", "2"}, "B": {"X"}}},
                {"A": {"1", "2"}, "B": {"X"}},
            )
        )

        self.assertFalse(
            _annotation_dict_all_filter(
                {ANNOTATIONS: {"A": {"1", "2"}, "B": {"X"}}},
                {"A": {"1", "2", "3"}, "B": {"X", "Y"}},
            )
        )

        self.assertFalse(_annotation_dict_all_filter({ANNOTATIONS: {"A": {"1"}}}, {"A": {"1", "2"}}))

        self.assertFalse(_annotation_dict_all_filter({ANNOTATIONS: {"A": {"1"}}}, {"A": {"2"}}))

        self.assertFalse(_annotation_dict_all_filter({ANNOTATIONS: {"A": {"1"}}}, {"B": {"1"}}))

    def test_all_filter_no_query(self):
        """Test that the all filter returns true when there's no argument"""
        graph = BELGraph()
        graph.add_increases(Protein(n(), n()), Protein(n(), n()), citation=n(), evidence=n())
        self.assertEqual(1, count_passed_edge_filter(graph, build_annotation_dict_all_filter({})))

    def test_all_filter_no_annotations(self):
        graph = BELGraph()
        graph.add_increases(Protein(n(), n()), Protein(n(), n()), citation=n(), evidence=n())
        self.assertEqual(
            0,
            count_passed_edge_filter(graph, build_annotation_dict_all_filter({"A": {"1"}})),
        )

    def test_all_filter_empty_annotations(self):
        graph = BELGraph()
        graph.add_increases(
            Protein(n(), n()),
            Protein(n(), n()),
            citation=n(),
            evidence=n(),
            annotations={},
        )
        self.assertEqual(
            0,
            count_passed_edge_filter(graph, build_annotation_dict_all_filter({"A": {"1"}})),
        )

    def test_all_filter(self):
        graph = BELGraph()
        graph.annotation_list["A"] = set("12345")
        graph.add_increases(
            Protein(n(), n()),
            Protein(n(), n()),
            citation=n(),
            evidence=n(),
            annotations={
                "A": {"1", "2", "3"},
            },
        )

        self.assertEqual(
            1,
            count_passed_edge_filter(
                graph,
                build_annotation_dict_all_filter(graph._clean_annotations({"A": {"1"}})),
            ),
        )
        self.assertEqual(
            1,
            count_passed_edge_filter(
                graph,
                build_annotation_dict_all_filter(graph._clean_annotations({"A": {"1", "2"}})),
            ),
        )
        self.assertEqual(
            1,
            count_passed_edge_filter(
                graph,
                build_annotation_dict_all_filter(graph._clean_annotations({"A": {"1", "2", "3"}})),
            ),
        )
        self.assertEqual(
            0,
            count_passed_edge_filter(
                graph,
                build_annotation_dict_all_filter(graph._clean_annotations({"A": {"1", "2", "3", "4"}})),
            ),
        )
        self.assertEqual(
            0,
            count_passed_edge_filter(
                graph,
                build_annotation_dict_all_filter(graph._clean_annotations({"A": {"4"}})),
            ),
        )

    def test_all_filter_dict(self):
        graph = BELGraph()
        graph.annotation_list["A"] = set("12345")
        a, b = Protein(namespace="hgnc", identifier="1", name="A"), Protein(namespace="hgnc", identifier="2", name="B")
        graph.add_increases(
            a,
            b,
            citation=n(),
            evidence=n(),
            annotations={
                "A": {"1", "2", "3"},
            },
        )

        self.assertEqual(
            1,
            count_passed_edge_filter(
                graph,
                build_annotation_dict_all_filter(graph._clean_annotations({"A": {"1": True}})),
            ),
        )
        self.assertEqual(
            1,
            count_passed_edge_filter(
                graph,
                build_annotation_dict_all_filter(graph._clean_annotations({"A": {"1": True, "2": True}})),
            ),
        )
        self.assertEqual(
            1,
            count_passed_edge_filter(
                graph,
                build_annotation_dict_all_filter(graph._clean_annotations({"A": {"1": True, "2": True, "3": True}})),
            ),
        )
        self.assertEqual(
            0,
            count_passed_edge_filter(
                graph,
                build_annotation_dict_all_filter(
                    graph._clean_annotations({"A": {"1": True, "2": True, "3": True, "4": True}})
                ),
            ),
        )
        self.assertEqual(
            0,
            count_passed_edge_filter(
                graph,
                build_annotation_dict_all_filter(graph._clean_annotations({"A": {"4": True}})),
            ),
        )



# -*- coding: utf-8 -*-

"""Misc node predicates."""

from typing import Iterable, Type

from .utils import node_predicate
from ..typing import NodePredicate
from ...graph import BELGraph
from ....constants import PART_OF, RELATION
from ....dsl import BaseEntity, ListAbundance

__all__ = [
    "none_of",
    "one_of",
    "is_isolated_list_abundance",
]


def none_of(nodes: Iterable[BaseEntity]) -> NodePredicate:
    """Build a node predicate that returns false for the given nodes."""
    nodes = set(nodes)

    @node_predicate
    def _predicate(node: BaseEntity) -> bool:
        """Return true if the node is not in the given set of nodes."""
        return node not in nodes

    return _predicate


def one_of(nodes: Iterable[BaseEntity]) -> NodePredicate:
    """Build a function that returns true for the given nodes."""
    nodes = set(nodes)

    @node_predicate
    def _predicate(node: BaseEntity) -> bool:
        """Return true if the node is in the given set of nodes."""
        return node in nodes

    return _predicate


def is_isolated_list_abundance(
    graph: BELGraph,
    node: BaseEntity,
    cls: Type[ListAbundance] = ListAbundance,
) -> bool:
    """Return if the node is a list abundance but has no qualified edges."""
    return (
        isinstance(node, cls)
        and 0 == graph.out_degree(node)
        and all(data[RELATION] == PART_OF for _, __, data in graph.in_edges(node, data=True))
    )



# -*- coding: utf-8 -*-

"""Tests for functions for grouping BEL graphs into sub-graphs."""

import unittest

from pybel import BELGraph
from pybel.constants import CITATION_TYPE_PUBMED
from pybel.dsl import protein
from pybel.language import Entity
from pybel.struct.grouping import get_subgraphs_by_annotation, get_subgraphs_by_citation
from pybel.testing.utils import n

test_namespace_url = n()
test_annotation_url = n()
citation, evidence = n(), n()
a, b, c, d = [protein(namespace="test", name=str(i)) for i in range(4)]


class TestAnnotation(unittest.TestCase):
    """Tests for getting sub-graphs by annotation."""

    def setUp(self):
        """Set up the test case with a pre-populated BEL graph."""
        self.graph = BELGraph()

        self.graph.namespace_url["test"] = test_namespace_url
        self.graph.annotation_url["subgraph"] = test_annotation_url

        self.graph.add_increases(
            a,
            b,
            citation=citation,
            evidence=evidence,
            annotations={"subgraph": {"1", "2"}},
        )
        self.graph.add_increases(a, c, citation=citation, evidence=evidence, annotations={"subgraph": {"1"}})
        self.graph.add_increases(
            b,
            d,
            citation=citation,
            evidence=evidence,
            annotations={"subgraph": {"1", "2"}},
        )
        self.graph.add_increases(a, d, citation=citation, evidence=evidence, annotations={"subgraph": {"2"}})
        self.graph.add_increases(c, d, citation=citation, evidence=evidence)

    def test_get_subgraphs_by_annotation(self):
        subgraphs = get_subgraphs_by_annotation(self.graph, annotation="subgraph")

        self.assertEqual(2, len(subgraphs))
        self.assertIn(Entity(namespace="subgraph", identifier="1"), subgraphs)
        self.assertIn(Entity(namespace="subgraph", identifier="2"), subgraphs)

        subgraph_1 = subgraphs[Entity(namespace="subgraph", identifier="1")]
        self.assertIsInstance(subgraph_1, BELGraph)

        self.assertIn("test", subgraph_1.namespace_url)
        self.assertIn("subgraph", subgraph_1.annotation_url)

        self.assertIn(a, subgraph_1)
        self.assertIn(b, subgraph_1)
        self.assertIn(c, subgraph_1)
        self.assertIn(d, subgraph_1)

        self.assertIn(b, subgraph_1[a])
        self.assertIn(c, subgraph_1[a])
        self.assertIn(d, subgraph_1[b])
        self.assertNotIn(d, subgraph_1[a])
        self.assertNotIn(d, subgraph_1[c])

        subgraph_2 = subgraphs[Entity(namespace="subgraph", identifier="2")]
        self.assertIsInstance(subgraph_2, BELGraph)

        self.assertIn("test", subgraph_2.namespace_url)
        self.assertIn("subgraph", subgraph_2.annotation_url)

        self.assertIn(a, subgraph_2)
        self.assertIn(b, subgraph_2)
        self.assertNotIn(c, subgraph_2)
        self.assertIn(d, subgraph_2)

        self.assertIn(b, subgraph_2[a])
        self.assertNotIn(c, subgraph_2[a])
        self.assertIn(d, subgraph_2[b])
        self.assertIn(d, subgraph_2[a])

    def test_get_subgraphs_by_annotation_with_sentinel(self):
        sentinel = n()
        subgraphs = get_subgraphs_by_annotation(self.graph, annotation="subgraph", sentinel=sentinel)

        self.assertEqual(3, len(subgraphs))
        self.assertIn(Entity(namespace="subgraph", identifier="1"), subgraphs)
        self.assertIn(Entity(namespace="subgraph", identifier="2"), subgraphs)
        self.assertIn(sentinel, subgraphs)


class TestProvenance(unittest.TestCase):
    """Tests for getting sub-graphs by provenance information (citation, etc.)"""

    def test_get_subgraphs_by_citation(self):
        """Test getting sub-graphs by citation."""
        graph = BELGraph()

        c1, c2, c3 = n(), n(), n()

        graph.add_increases(a, b, citation=c1, evidence=n())
        graph.add_increases(a, b, citation=c2, evidence=n())
        graph.add_increases(b, c, citation=c1, evidence=n())
        graph.add_increases(c, d, citation=c1, evidence=n())
        graph.add_increases(a, d, citation=c3, evidence=n())

        subgraphs = get_subgraphs_by_citation(graph)

        # TODO tests for metadata

        c1_pair = (CITATION_TYPE_PUBMED, c1)
        self.assertIn(c1_pair, subgraphs)
        c1_subgraph = subgraphs[c1_pair]
        self.assertIn(a, c1_subgraph)
        self.assertIn(b, c1_subgraph)
        self.assertIn(c, c1_subgraph)
        self.assertIn(d, c1_subgraph)

        c2_pair = (CITATION_TYPE_PUBMED, c2)
        self.assertIn(c2_pair, subgraphs)
        c2_subgraph = subgraphs[c2_pair]
        self.assertIn(a, c2_subgraph)
        self.assertIn(b, c2_subgraph)
        self.assertNotIn(c, c2_subgraph)
        self.assertNotIn(d, c2_subgraph)

        c3_pair = (CITATION_TYPE_PUBMED, c3)
        self.assertIn(c3_pair, subgraphs)
        c3_subgraph = subgraphs[c3_pair]
        self.assertIn(a, c3_subgraph)
        self.assertNotIn(b, c3_subgraph)
        self.assertNotIn(c, c3_subgraph)
        self.assertIn(d, c3_subgraph)



# -*- coding: utf-8 -*-

"""Warnings for old TSV conversion module."""

import warnings

from ..triples import to_edgelist
from ..triples import to_triples_file as to_tsv

__all__ = [
    "to_tsv",
    "to_edgelist",
]

warnings.warn(
    """Use pybel.io.triples module instead. Changes in PyBEL v0.15.0:

  - pybel.to_tsv renamed to pybel.to_triples_file

Will be removed in PyBEL v0.16.*
  """,
    DeprecationWarning,
)



# -*- coding: utf-8 -*-

"""Connection configuration for PyBEL."""

import logging

import pystow

__all__ = [
    "connection",
    "PYBEL_MINIMUM_IMPORT_VERSION",
    "PYBEL_HOME",
]

logger = logging.getLogger(__name__)

#: The last PyBEL version where the graph data definition changed
PYBEL_MINIMUM_IMPORT_VERSION = 0, 14, 0

PYBEL_HOME = pystow.join("pybel")
DEFAULT_CACHE_NAME = "pybel_{}.{}.{}_cache.db".format(*PYBEL_MINIMUM_IMPORT_VERSION)
DEFAULT_CACHE_PATH = pystow.join("pybel", name=DEFAULT_CACHE_NAME)
#: The default cache connection string uses sqlite.
DEFAULT_CACHE_CONNECTION = "sqlite:///" + DEFAULT_CACHE_PATH.as_posix()

connection = pystow.get_config(
    "pybel",
    "connection",
    default=DEFAULT_CACHE_CONNECTION,
)
